# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
 
import datetime 
from abc import abstractmethod
from dhi.platform.base.client import DataContract
from dhi.platform.base.exceptions import MikeCloudException
from .base import constants
from enum import Enum
from typing import Tuple
from .generated.metadatagen import UnitIdV3, ItemIdV3


UnitId = UnitIdV3

ItemId = ItemIdV3

class SpatialFilter(DataContract):
    def __init__(self, geometry:str, srid:int=None) -> None:
        """
        SpatialFilter for transfer/conversion spatial filter transformation.

        :param geometry: WKT geometry (polygon) to filter by
        :param srid: Spatial Reference ID of the geometry if different from the dataset SRID
        """
        self._geometry = geometry
        self._srid = srid
    
    def body(self) -> dict:
        body = {
            "geometry": self._geometry,
        }
        if self._srid:
            body["SRID"] = self._srid
        return body

    def to_dict(self) -> dict:
        return self.body()


class TemporalFilter(DataContract):
    pass

    @abstractmethod
    def body(self) -> dict:
        return {}

    @abstractmethod
    def to_dict(self) -> dict:
        return self.body()
    
    @classmethod
    def create_index_filter(cls, from_:int=None, to:int=None):
        return TemporalIndexFilter(from_, to)

    @classmethod
    def create_value_filter(cls, from_:datetime.datetime, to:datetime.datetime):
        return TemporalValueFilter(from_, to)

    @classmethod
    def create_index_list_filter(cls, indices:Tuple[int]):
        return TemporalIndexListFilter(indices)


class TemporalIndexFilter(TemporalFilter):
    def __init__(self, from_:int=None, to:int=None) -> None:
        super().__init__()
        if not (from_ or to):
            raise MikeCloudException("At least one of from_ or to parameters must be specified")
        self._from = from_
        self._to = to

    def body(self) -> dict:
        body = { "type": "TemporalIndexFilter" }

        if self._from is not None:
            body["from"] = self._from

        if self._to is not None:
            body["to"] = self._to
        
        return body

        
class TemporalValueFilter(TemporalFilter):
    def __init__(self, from_:datetime.datetime, to:datetime.datetime) -> None:
        super().__init__()
        if not (from_ or to):
            raise MikeCloudException("At least one of from_ or to parameters must be specified")
        self._from = from_
        self._to = to
    
    def body(self) -> dict:
        body = { "type": "TemporalValueFilter" }

        if self._from:
            body["from"] = self._from.strftime(constants.DATETIMEFORMAT)
        if self._to:
            body["to"] = self._to.strftime(constants.DATETIMEFORMAT)

        return body

        
class TemporalIndexListFilter(TemporalFilter):
    def __init__(self, indices:Tuple[int]) -> None:
        super().__init__()
        self._indices = indices
    
    def body(self) -> dict:
        return { 
            "type": "TemporalIndexListFilter",
            "indices": self._indices 
        }


class VerticalFilter(DataContract):
    pass 

    @abstractmethod
    def body(self) -> dict:
        return {}

    @abstractmethod
    def to_dict(self) -> dict:
        return self.body()

    @classmethod
    def create_index_filter(from_:int=None, to:int=None):
        return VerticalIndexFilter(from_, to)


class VerticalIndexFilter(VerticalFilter):
    def __init__(self, from_:int=None, to:int=None) -> None:
        if not (from_ or to):
            raise MikeCloudException("At least one of from_ or to parameters must be specified")
        self._from = from_
        self._to = to

    def body(self) -> dict:
        body = {}
        
        if self._from:
            body["from"] = self._from
        if self._to:
            body["to"] = self._to

        return body


class AttributeDataType(Enum):
    TEXT = "Text"
    DATE = "Date"
    INT32 = "Int32"
    INT64 = "Int64"
    SINGLE = "Single"
    DOUBLE = "Double"
    INT16 = "Int16"

    @property
    def name(self):
        return self.value

    @classmethod
    def from_string(cls, x:str):
        return {
            "text": AttributeDataType.TEXT,
            "date": AttributeDataType.DATE,
            "int32": AttributeDataType.INT32,
            "int64": AttributeDataType.INT64,
            "single": AttributeDataType.SINGLE,
            "double": AttributeDataType.DOUBLE,
            "int16": AttributeDataType.INT16
        }[x.lower()]


class TimeSeriesDataType(Enum):
    INSTANTANEOUS = "Instantaneous"
    ACCUMULATED = "Accumulated"
    STEP_ACCUMULATED = "StepAccumulated"
    MEAN_STEP_BACKWARD = "MeanStepBackward"
    MEAN_STEP_FORWARD = "MeanStepForward"

    @property
    def name(self):
        return self.value

    @classmethod
    def from_string(cls, x:str):
        return {
            "instantaneous": TimeSeriesDataType.INSTANTANEOUS,
            "accumulated": TimeSeriesDataType.ACCUMULATED,
            "stepaccumulated":TimeSeriesDataType.STEP_ACCUMULATED,
            "meanstepbackward": TimeSeriesDataType.MEAN_STEP_BACKWARD,
            "meanstepforward": TimeSeriesDataType.MEAN_STEP_FORWARD
        }[x.lower()]



class ItemDefinition:
    
    def __init__(
        self,
        name:str,
        unit:str,
        item:str,
        data_type:AttributeDataType,
        timeseries_data_type: TimeSeriesDataType=None,
        no_data_value:float=None,
        has_layers:bool=False
    ) -> None:
        self._name = name
        self._unit = unit
        self._item = item
        self._data_type = data_type
        self._timeseries_data_type = timeseries_data_type
        self._no_data_value = no_data_value
        self._has_layers = has_layers
    
    @property
    def name(self) -> str:
        return self._name
    
    @property
    def unit(self):
        return self._unit
    
    @property
    def item(self):
        return self._item

    @property
    def data_type(self):
        return self._data_type
        
    @property
    def timeseries_data_type(self):
        return self._timeseries_data_type
    
    @property
    def no_data_value(self):
        return self._no_data_value
    
    @property
    def has_layers(self) -> bool:
        return self._has_layers

    def body(self) -> dict:
        body = {
            "name": self._name,
            "unit": self._unit,
            "item": self._item,
            "dataType": self._data_type.name
        }
        
        if self._no_data_value is not None:
            body["noDataValue"] = self._no_data_value
        
        if self._has_layers is not None:
            body["hasLayers"] = self._has_layers
        
        if self._timeseries_data_type:
            body["timeSeriesType"] = self._timeseries_data_type.name
        
        return body

    @classmethod
    def from_body(cls, body:dict):
        ts_type = body.get("timeSeriesType", None)
        ts_data_type = None if ts_type is None else TimeSeriesDataType.from_string(ts_type)

        return cls(
            name = body["name"],
            unit = body["unit"],
            item = body["item"],
            data_type = AttributeDataType.from_string(body["dataType"]),
            timeseries_data_type = ts_data_type,
            no_data_value = body.get("noDataValue", None),
            has_layers = bool(body.get("hasLayers", False))
        )


class PropertyDataType(Enum):
    DATETIME = "DateTime"
    LONG = "Long"
    DOUBLE = "Double"
    BOOLEAN = "Boolean"
    TEXT = "Text"

    @classmethod
    def from_string(cls, x:str):
        return {
            "datetime": PropertyDataType.DATETIME,
            "long": PropertyDataType.LONG,
            "double": PropertyDataType.DOUBLE,
            "boolean": PropertyDataType.BOOLEAN,
            "text": PropertyDataType.TEXT,
        }[x.lower()]

    @property
    def name(self):
        return self.value


class PropertyDefinition:

    def __init__(self, name, data_type:PropertyDataType) -> None:
        self._name = name
        self._data_type = data_type

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def data_type(self) -> PropertyDataType:
        return self._data_type

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            name=body["name"],
            data_type=PropertyDataType.from_string(body["dataType"])
        )
    
    def body(self) -> dict:
        return { "name": self._name, "dataType": self._data_type.name }


class AttributeOperator(Enum):
    EQUAL = "Equal"
    NOT_EQUAL = "NotEqual"
    GREATER_THAN = "GreaterThan"
    LESS_THAN = "LessThan"
    GREATER_THAN_OR_EQUAL = "GreaterThanOrEqual"
    LESS_THAN_OR_EQUAL = "LessThanOrEqual"
    CONTAINS = "Contains"
    STARTS_WITH = "StartsWith"

    @property
    def name(self):
        return self.value


class DatasetType(Enum):
    FILE = "file"
    MULTIDIMENSIONAL = "multidimensional"
    TIMESERIES = "timeseries"
    GISVECTORDATA = "gisvectordata"
    TILES = "tiles"

    @property
    def name(self) -> str:
        return self.value


class SpatialOperator(Enum):
    WITHIN = "Within"
    INTERSECTS = "Intersects"

    @property
    def name(self):
        return self.value


class ComparisonOperator(Enum):
    EQUAL = "Equal"
    NOT_EQUAL = "NotEqual"
    GREATER_THAN = "GreaterThan"
    LESS_THAN = "LessThan"
    GREATER_THAN_OR_EQUAL = "GreaterThanOrEqual"
    LESS_THAN_OR_EQUAL = "LessThanOrEqual"

    @property
    def name(self):
        return self.value


class QueryCondition:

    def __init__(self, type:str):
        self._type = type
    
    @classmethod
    def create_attribute_query_condition(self, name, operator, value):
        return AttributeQueryCondition(name, operator, value)

    @classmethod
    def create_dataset_type_query_condition(self, dataset_type:DatasetType):
        return DatasetTypeQueryCondition(dataset_type)

    @classmethod
    def create_fultlext_query_condition(self, search_string):
        return FullTextQueryCondition(search_string)

    @classmethod
    def create_ids_query_condition(self, ids=Tuple[object]):
        return IdsQueryCondition(ids)

    @classmethod
    def create_spatial_query_condition(self, geometry, operator:SpatialOperator=SpatialOperator.WITHIN):
        return SpatialQueryCondition(geometry, operator)
    
    @classmethod
    def create_created_at_query_condition(self, created_at:datetime.datetime, operator:ComparisonOperator=ComparisonOperator.EQUAL):
        return CreatedAtQueryCondition(created_at, operator)
    
    @classmethod
    def create_catalog_query_condition(self, catalog_id):
        return CatalogIdQueryCondition(catalog_id)

    def body(self) -> dict:
        return { "type": self._type }


class AttributeQueryCondition(QueryCondition):

    def __init__(self, name:str, operator:AttributeOperator, value):
        super().__init__(type(self).__name__)
        self._name = name
        self._operator = operator
        self._value = value
    
    def body(self) -> dict:
        body = super().body()
        body.update({
            "name": self._name,
            "operator": self._operator.name,
            "value": self._value
        })
        return body


class DatasetTypeQueryCondition(QueryCondition):

    def __init__(self, dataset_type:DatasetType):
        super().__init__(type(self).__name__)
        self._dataset_type = dataset_type

    def body(self) -> dict:
        body = super().body()
        body.update({
            "datasetType": self._dataset_type.name
        })
        return body


class FullTextQueryCondition(QueryCondition):

    def __init__(self, search_string:str):
        super().__init__(type(self).__name__)
        self._search_string = search_string

    def body(self) -> dict:
        body = super().body()
        body.update({
            "searchString": self._search_string
        })
        return body


class IdsQueryCondition(QueryCondition):

    def __init__(self, ids=Tuple[object]):
        super().__init__(type(self).__name__)
        self._ids = ids

    def body(self) -> dict:
        body = super().body()
        body.update({
            "ids": self._ids
        })
        return body


class SpatialQueryCondition(QueryCondition):

    def __init__(self, geometry:object, operator:SpatialOperator=SpatialOperator.WITHIN):
        super().__init__(type(self).__name__)
        self._geometry = geometry
        self._operator = operator

    def body(self) -> dict:
        body = super().body()
        body.update({
            "geometry": self._geometry,
            "operator": self._operator.name
        })
        return body


class CreatedAtQueryCondition(QueryCondition):

    def __init__(self, created_at:datetime.datetime, operator:ComparisonOperator=ComparisonOperator.EQUAL):
        super().__init__(type(self).__name__)
        self._created_at = created_at
        self._operator = operator

    def body(self) -> dict:
        body = super().body()
        body.update({
            "createdAt": self._created_at.strftime(constants.DATETIMEFORMAT),
            "operator": self._operator.name
        })
        return body


class CatalogIdQueryCondition(QueryCondition):

    def __init__(self, catalog_id):
        super().__init__(type(self).__name__)
        self._catalog_id = catalog_id

    def body(self) -> dict:
        body = super().body()
        body.update({
            "catalogId": self._catalog_id
        })
        return body


import unittest
from unittest.mock import ANY, create_autospec, Mock
import datetime
from unittest.case import _AssertRaisesContext
import uuid
from dhi.platform import metadata, transfer
from dhi.platform.authentication import ApiKeyIdentity
from dhi.platform.base.client import Response
from dhi.platform.raw import RawClientV2
import os

class TestTransferTest(unittest.TestCase):

    _verbosity = 0
    _project_id = None
    _identity = None
    _test_data_dir = None
    
    def setUp(self) -> None:
        if not self._identity:
            self._identity = ApiKeyIdentity(str(uuid.uuid4()), str(uuid.uuid4()))
        self._metadata2 = create_autospec(metadata.MetadataClientV2(verbose=self._verbosity, identity=self._identity))
        self._metadata3 = create_autospec(metadata.MetadataClientV3(verbose=self._verbosity, identity=self._identity))
        self._raw2 = create_autospec(RawClientV2(verbose=self._verbosity, identity=self._identity))

        self._transferclient = transfer.TransferClient(
            verbose=self._verbosity, 
            identity=self._identity,
            MetadataClientV2=self._metadata2,
            MetadataClientV3=self._metadata3,
            RawClientV2=self._raw2)
        
        self._test_data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
        self._project_id = str(uuid.uuid4())

    def tearDown(self) -> None:
        super().tearDown()
        
    def test_get_readers_is_ok(self):
        readers = [*self._transferclient.get_readers()]
        self._metadata2.GetReadersListV2.assert_called_once()

    def test_simple_file_upload_is_ok(self):

        self._metadata3.UploadConvertV3 = Mock(return_value=Response(1, 200, "", {}, body={
            "id": str(uuid.uuid4()),
            "createdAt": "2022-01-01T01:01:01",
            "createdBy": str(uuid.uuid4()),
            "type": "Import",
            "format": None,
            "status": "Completed"
        }))

        destination_project_id = str(uuid.uuid4())
        self._transferclient.create_url_import(destination_project_id, "http://url", "file") \
            .execute()

        self._metadata3.UploadConvertV3.assert_called_once_with({
            "originalFileName": "file",
            "uploadUrl": "http://url",
            "outputDatasetData": { "name": "file" },
            "readerName": "FileReader",
            "writerName": "FileWriter",
            "projectId": destination_project_id
        })

    def test_upload_with_parameters_is_ok(self):

        self._metadata3.UploadConvertV3 = Mock(return_value=Response(1, 200, "", {}, body={
            "id": str(uuid.uuid4()),
            "createdAt": "2022-01-01T01:01:01",
            "createdBy": str(uuid.uuid4()),
            "type": "Import",
            "format": None,
            "status": "Completed"
        }))

        destination_project_id = str(uuid.uuid4())
        reader = transfer.DfsuReader()
        reader.with_allowed_items(("item1", "item2"))

        self._transferclient.create_url_import(destination_project_id, "http://url", "dataset", "file") \
            .with_reader(reader) \
            .with_writer(transfer.MDWriter()) \
            .execute()
        
        self._metadata3.UploadConvertV3.assert_called_once_with({
            "originalFileName": "file",
            "uploadUrl": "http://url",
            "outputDatasetData": { "name": "dataset" },
            "projectId": destination_project_id,
            "readerName": "DfsuReader",
            "readerParameters": [{"name": "AllowedItemNames", "value": ["item1", "item2"]}],
            "writerName": "MDWriter"
        })
    
    def test_get_dataset_transfer_list_is_ok(self):
        project_id = str(uuid.uuid4())
        dataset_id = str(uuid.uuid4())
        date = datetime.date.today()

        self._metadata2.GetProjectTransferListV2 = Mock(return_value=Response(1, 200, "", {}, body={
            "data": []
        }))

        transfers = [*self._transferclient.list_dataset_transfers(project_id, dataset_id, from_=date)]

        self.assertEqual(transfers, [])

        self._metadata2.GetProjectTransferListV2.assert_called_once_with(
            projectid=project_id,
            datasetid=dataset_id,
            from_=date,
            to=None,
            status=(),
            offset=ANY,
            limit=ANY
        )

    def test_get_project_transfer_list_is_ok(self):
        project_id = str(uuid.uuid4())
        date = datetime.date.today()

        self._metadata2.GetProjectTransferListV2 = Mock(return_value=Response(1, 200, "", {}, body={
            "data": []
        }))

        transfers = [*self._transferclient.list_project_transfers(project_id, from_=date)]

        self.assertEqual(transfers, [])

        self._metadata2.GetProjectTransferListV2.assert_called_once_with(
            project_id,
            datasetid=None,
            from_=date,
            to=None,
            status=(),
            offset=ANY,
            limit=ANY
        )

    def test_is_dataset_id_locked_is_ok(self):
        project_id = str(uuid.uuid4())
        dataset_id = str(uuid.uuid4())

        self._metadata2.GetProjectTransferListV2 = Mock(return_value=Response(1, 200, "", {}, body={
            "data": [
                {
                    "id": str(uuid.uuid4()),
                    "createdAt": "2022-01-01T01:01:01",
                    "createdBy": str(uuid.uuid4()),
                    "type": "Import",
                    "format": "file",
                    "status": "InProgress"
                }
            ]
        }))
        
        is_locked = self._transferclient.is_dataset_locked(project_id, dataset_id)
        
        self.assertTrue(is_locked)

        self._metadata2.GetProjectTransferListV2.assert_called_once_with(
            projectid=project_id,
            datasetid=dataset_id,
            from_=None,
            to=None,
            status=(transfer.TransferStatus.NONE, transfer.TransferStatus.PENDING, transfer.TransferStatus.INPROGRESS),
            offset=ANY,
            limit=ANY
        )

    def test_upload_staged_files_is_ok(self):
        project_id = str(uuid.uuid4())
        upload_input = transfer.StagedFilesUploadInput(files=[], create_destination_path_if_not_exists=True)
        expected = transfer.StagedFilesUploadOutput(datasets=[], failures=[])

        self._metadata2.UploadStagedFilesV2 = Mock(return_value=Response(1, 200, "", {}, body={
            "datasets": [],
            "failures": []
        }))

        result = self._transferclient.upload_staged_files(project_id, upload_input)

        self.assertEqual(result.datasets, expected.datasets)
        self.assertEqual(result.failures, expected.failures)
        

if __name__ == "__main__":
    unittest.main()


# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json, os, sys, urllib.parse, uuid
from . import constants
from json.decoder import JSONDecodeError
from abc import ABC
from http.client import HTTPConnection, HTTPSConnection
from dhi.platform.authentication import MikeCloudIdentity
from .exceptions import MikeCloudException, MikeCloudRestApiException
from typing import Any, Dict, get_type_hints, get_origin, get_args
from enum import Enum
#from .fmt import Format
#import sys
#from argparse import ArgumentParser
#from http.client import HTTPConnection, HTTPSConnection
#from os import environ

class Response:
    def __init__(self, operationid, status, reason, headers, body=None, requestheaders=None, requestBody=None):
        self.__operationId = operationid
        self.__status = status
        self.__reason = reason
        self.__headers = headers
        self.__body = body
        self.__requestHeaders = requestheaders
        self.__requestBody = requestBody

    def Dump(self, file=sys.stdout):
        print(f"{self.Status} {self.Reason}", file=file)
        if self.Body:
            json.dump(self.Body, file, indent=True)

    @property
    def OperationId(self):
        return self.__operationId

    @property
    def Status(self):
        return self.__status

    @property
    def Reason(self):
        return self.__reason

    @property
    def Headers(self):
        return self.__headers

    @property
    def Body(self):
        return self.__body

    @property
    def RequestHeaders(self):
        return self.__requestHeaders

    @property
    def RequestBody(self):
        return self.__requestBody

    @property
    def IsOk(self):
        return self.Status >= 200 and self.Status < 300

    @property
    def IsError(self):
        return not self.IsOk


class Contracts:
    @staticmethod
    def SetBodyField(body, field, value, default=None):
        if value != None:
            body[field] = value
            return value
        if default != None and not field in body:
            body[field] = default
            return default
        return body.get(field)

    @classmethod
    def PrepareField(cls, body, fieldname, default):
        return cls.SetBodyField(body, fieldname, None, default)


class PlatformClientBase(ABC):
    def __init__(self, inspectFnc, verbose=0, **kwargs):
        self.__verbose = verbose
        self.__inspectFnc = inspectFnc

    def _Inspect(self, operationid, level, what, msg=None, obj=None):
        if self.__inspectFnc:
            self.__inspectFnc(operationid, self.__verbose, level, what, msg, obj)

    @property
    def Verbose(self):
        return self.__verbose

    @property
    def NextOperationId(self):
        return str(uuid.uuid1())

    @staticmethod
    def DefaultInspectFnc(operationid, verbose, level, what, msg, obj):
        if verbose and verbose >= level:
            if what:
                print(f"### {what}")
            if msg:
                print(f"{msg}")
            if obj:
                json.dump(obj, sys.stdout, indent=True)
                print("")


class DummyAuthenticator:
    def UpdateHeaders(self, headers):
        pass


class OpenApiKeyAuthenticator:
    def __init__(self, apikey):
        self.__apikey = apikey

    def UpdateHeaders(self, headers):
        if self.__apikey:
            headers["dhi-open-api-key"] = self.__apikey


class BearerAuthenticator:
    def __init__(self, accesstoken, customerid=None, userid=None):
        self.__accesstoken = accesstoken
        self.__customerid = customerid
        self.__userid = userid

    def UpdateHeaders(self, headers):
        if self.__accesstoken:
            headers["Authorization"] = f"Bearer {self.__accesstoken}"
        if self.__customerid:
            headers["dhi-customer-guid"] = self.__customerid
        if self.__userid:
            headers["dhi-user-id"] = self.__userid


class LocalAuthenticator:
    def __init__(self, customerid=None, userid=None, userisadmin=None):
        self.__customerid = customerid
        self.__userid = userid
        self.__userisadmin = userisadmin

    def UpdateHeaders(self, headers):
        if self.__customerid:
            headers["dhi-customer-guid"] = self.__customerid
        if self.__userid:
            headers["dhi-user-id"] = self.__userid
        if self.__userisadmin:
            headers["dhi-user-is-admin"] = self.__userisadmin


class PlatformClient(PlatformClientBase):
    def __init__(self, inspectFnc, includeheaders=None, **kwargs):
        super().__init__(inspectFnc, **kwargs)

        if identity := kwargs.get("identity"):
            if not isinstance(identity, MikeCloudIdentity):
                 raise MikeCloudException("Invalid identity provided")
            kwargs.update(identity.get_parameters())
            kwargs.pop("identity")
            
        self.__InitCoreMetadataUrl(**kwargs)
        #self.__basicHeaders = {"Content-Type": "application/json; charset=UTF-8"}
        self.__basicHeaders = {"Content-Type": "application/json-patch+json", "Accept": "text/plain", "dhi-pythonsdk-version": constants.DHI_PYTHONSDK_VERSION}
        if includeheaders:
            self.__basicHeaders.update(includeheaders)
        self.__authenticator = self.__CreateAuthenticator(**kwargs)
        self._Inspect(self.NextOperationId, 3, "kwargs", None, kwargs)

    def __CreateAuthenticator(self, apikey=None, accesstoken=None, preferapikey=False, customerid=None, userid=None, userisadmin=None, **kwargs):
        if self.IsLocal:
            return LocalAuthenticator(customerid, userid, userisadmin)
        if accesstoken and (not apikey or not preferapikey):
            return BearerAuthenticator(accesstoken, customerid, userid)
        if apikey:
            return OpenApiKeyAuthenticator(apikey)
        return DummyAuthenticator()

    def GetHeaders(self, includeheaders=None, **kwargs):
        return self.GetBasicHeaders(includeheaders, **kwargs)

    @classmethod
    def GetQueryParams(cls, **kwargs):
        params = []
        for name, value in kwargs.items():
            if value:
                #key.capitalize() if capitalize else key
                if isinstance(value, list):
                    params.extend((cls.__FormatParam(name, x) for x in value))
                else:
                    params.append(cls.__FormatParam(name, value))
        return params

    def GetRequest(self, url, queryparams=None, includeheaders=None, **kwargs) -> Response:
        return self.__Request("GET", url, None, queryparams, includeheaders, **kwargs)

    def DeleteRequest(self, url, queryparams=None, includeheaders=None, **kwargs) -> Response:
        return self.__Request("DELETE", url, None, queryparams, includeheaders, **kwargs)

    def PostRequest(self, url, body, queryparams=None, includeheaders=None, **kwargs) -> Response:
        return self.__Request("POST", url, body, queryparams, includeheaders, **kwargs)

    def PutRequest(self, url, body, queryparams=None, includeheaders=None, **kwargs) -> Response:
        return self.__Request("PUT", url, body, queryparams, includeheaders, **kwargs)

    def PatchRequest(self, url, body, queryparams=None, includeheaders=None, **kwargs) -> Response:
        return self.__Request("PATCH", url, body, queryparams, includeheaders, **kwargs)

    @property
    def IsLocal(self):
        return self.__coremetadataaddr.startswith("localhost") or self.__coremetadataaddr.startswith("127.0.0.1")

    @staticmethod
    def GetServiceHeaders(service, headers=None):
        result = headers if headers else {}
        if service:
            result["dhi-service-id"] = service
        return result

    @staticmethod
    def GetVersionHeaders(version, headers=None):
        result = headers if headers else {}
        if version:
            result["api-version"] = version
        return result

    def GetBasicHeaders(self, includeheaders=None, api_version=None, content_type=None, **kwargs):
        #headers = {"dhi-open-api-key": apikey, "dhi-project-id": projectid, "dhi-service-id": "engine"}
        headers = self.__basicHeaders.copy()
        self.__authenticator.UpdateHeaders(headers)
        if api_version:
            headers["api-version"] = api_version
        if projectid := kwargs.get("projectid"):
            headers["dhi-project-id"] = projectid
        if datasetid := kwargs.get("datasetid"):
            headers["dhi-dataset-id"] = datasetid
        if recursivetoken := kwargs.get("recursivetoken"):
            headers["dhi-recursive-token"] = recursivetoken
        if content_type:
            headers["Content-Type"] = content_type
        if includeheaders:
            headers.update(includeheaders)
        return headers

    def GetBasicHeadersV2(self, **kwargs):
        return self.GetBasicHeaders(api_version="2", **kwargs)

    def GetBasicHeadersV3(self, **kwargs):
        return self.GetBasicHeaders(api_version="3", **kwargs)

    @staticmethod
    def __FormatParam(name, value):
        return f"{name}={urllib.parse.quote_plus(value) if isinstance(value, str) else value}"

    def __GetConnection(self, operationid):
        if self.__coremetadatasecure:
            self._Inspect(operationid, 1, "connection", f"https://{self.__coremetadataaddr}")
            return HTTPSConnection(self.__coremetadataaddr)
        else:
            self._Inspect(operationid, 1, "connection", f"http://{self.__coremetadataaddr}")
            return HTTPConnection(self.__coremetadataaddr)
        #self._Inspect(operationid, 1, "connection", "{0}://{1}".format(
        #    "HTTPS" if self.__coremetadatasecure else "HTTP", self.__coremetadataaddr))
        #return HTTPSConnection(self.__coremetadataaddr) if self.__coremetadatasecure else HTTPConnection(self.__coremetadataaddr)

    def __Request(self, method, url, body=None, queryparams=None, includeheaders=None, **kwargs) -> Response:
        try:
            self._Inspect(operationid := self.NextOperationId, 1, "{{")
            conn = self.__GetConnection(operationid)
            if queryparams:
                url = f"{url}{'?'}{'&'.join(queryparams)}"
            self._Inspect(operationid, 1, "request", f"{method} {url}")
            headers = self.GetHeaders(includeheaders, **kwargs)
            if headers:
                self._Inspect(operationid, 2, "headers", None, headers)
            if body:
                self._Inspect(operationid, 2, "requestbody", None, body)
            
            if isinstance(body, str):
                conn.request(method, url, body if body else None, headers=headers)
            else:
                conn.request(method, url, json.dumps(body) if body else None, headers=headers)

            response = conn.getresponse()
            self._Inspect(operationid, 1, "response", f"{response.status} {response.reason}")
            self._Inspect(operationid, 2, "responseheaders", None, dict(response.headers))
            responsestr = response.read()
            conn.close()            
            if responsestr:
                self._Inspect(operationid, 1, "responsebodysize", f"{len(responsestr)}")
                self._Inspect(operationid, 3, "responsebody", responsestr)
                if response.headers.get("Content-Type", "application/json") == "application/x-binary":
                    responsejson = responsestr
                else:
                    responsejson = json.loads(responsestr)
                    self._Inspect(operationid, 2, "responsebodyjson", None, responsejson)
            self.__Raise_if_response_not_ok(response, responsestr)
            if responsestr:
                return Response(operationid, response.status, response.reason, response.headers, responsejson, headers, body)
            return Response(operationid, response.status, response.reason, dict(response.headers), None, headers, body)
        finally:
            self._Inspect(operationid, 1, "}}")

    def __InitCoreMetadataUrl(self, **kwargs):
        if self.__InitUrlAddressAndSecure(**kwargs):
            return True
        else:
            platformenv = self.__GetEnvironment("prod", **kwargs)
            if platformenv and platformenv.lower().endswith("eu"):
                tmp = platformenv.lower()[:-2]
                self.__coremetadataaddr = f"metadata-mike-platform-{tmp}.eu.mike-cloud-{tmp}.com"
                self.__coremetadatasecure = True
            elif platformenv and platformenv == "prod":
                #self.__coremetadataaddr = "metadata-mike-platform-prod.eu.mike-cloud.com"
                self.__coremetadataaddr = "api.mike-cloud.com"
                self.__coremetadatasecure = True
            elif platformenv and platformenv != "local":
                self.__coremetadataaddr = f"api.mike-cloud-{platformenv.lower()}.com"
                self.__coremetadatasecure = True
            else:
                self.__coremetadataaddr = "localhost:51499"
                self.__coremetadatasecure = False

    def __GetEnvironment(self, defaultenvironment, environment=None, **kwargs):
        return environment if environment else defaultenvironment

    def __InitUrlAddressAndSecure(self, metaurl=None, **kwargs):
        if metaurl:
            url = urllib.parse.urlparse(metaurl)
            self.__coremetadataaddr = f"{url.hostname}:{url.port}" if url.port else url.hostname
            self.__coremetadatasecure = url.scheme == "https"
            return True
        return False
    
    def __Raise_if_response_not_ok(self, response, responsestr):
        status = response.status
        if status <200 or status>=400:
            message = f"REST API call failed with status code {response.status}"
            detail = ""
            try:
                if responsestr:
                    problemdetails = json.loads(responsestr)
                    # Problem Details come out with Capitalized keys, this line fixes it:
                    problemdetails = dict(map(lambda k:(k.lower(),problemdetails[k]),problemdetails.keys()))

                    message = problemdetails.get("title", message)
                    detail = problemdetails.get("detail", detail)
                    type = problemdetails.get("type", None)
                    instance = problemdetails.get("instance", None)
                    errors = problemdetails.get("errors", None)

                    if errors:
                        try:
                            detail += json.dumps(errors)
                        except:
                            detail += str(errors) 

                    if detail and not message:
                        message = detail

                    raise MikeCloudRestApiException(message, detail, status, type, instance)
                else:
                    message = "Response body is empty"
                    detail = "Problem Details are not available"
                    raise MikeCloudRestApiException(message, detail, status, responsestr)
            except JSONDecodeError as er:
                raise MikeCloudRestApiException(message, detail, status, responsestr, er)

class DataContract(object):
    def to_dict(self) -> Dict[str, Any]:
        return dict()

    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        pass

    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        return dict()

    def get_dictionary(self, renamed: Dict[str, str]) -> Dict[str, Any]:
        result = vars(self)
        for name, value in result.items():
            if isinstance(value, list):
                result[name] = [x.to_dict() if isinstance(x, DataContract) else x for x in value]
            elif isinstance(value, DataContract):
                result[name] = value.to_dict()
        for o, n in renamed.items():
            if n in result:
                result[o] = result.pop(n)
        return result

    @staticmethod
    def load_from_directory(obj, src_dict: Dict[str, Any], renamed) -> None:
        hints = get_type_hints(type(obj))
        for k, v in src_dict.items():
            if k in renamed:
                k = renamed[k]
            h = hints.get(k)
            if isinstance(h, type):
                if issubclass(h, DataContract):
                    v = h.from_dict(v)
                elif issubclass(h, Enum):
                    v = h(v)
            elif get_origin(h) == list:
                hargs = get_args(h)
                hargstype = hargs[0]
                if issubclass(hargstype, DataContract):
                    v = [hargstype.from_dict(x) for x in v]
                elif issubclass(hargstype, Enum):
                    v = [hargstype(x) for x in v]
            setattr(obj, k, v)

    @staticmethod
    def __has_method(obj, name) -> bool:
        m = getattr(obj, name, None)
        return callable(m)

if __name__ == '__main__':
    print(__file__)
    print(dir())



# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#from _typeshed import Self
from enum import Enum
import datetime
import time
from typing import Generator, List, Tuple
from dhi.platform.base.exceptions import MikeCloudException, MikeCloudRestApiException
from .generated.metadatagen import AccessLevelV3, DatasetOutputV3, DatasetRecursiveListOutputV3, ItemIdV3, MetadataGenClientV1, MetadataGenClientV2, MetadataGenClientV3, PathActionCreateIfNotExistsV3, PathActionCreateV3, PathActionDeleteV3, PathActionV3, PrepareHierarchyInputV3, PrepareHierarchyOutputV3, UnitIdV3
from .base import constants
from .commonmodels import DatasetType


class AccessLevel(Enum):
    SHARED = 100
    PRIVATE = 200
    CONFIDENTIAL = 300

    @classmethod
    def from_string(self, access_level: str):
        options = {
            'Shared': self.SHARED,
            'Private': self.PRIVATE,
            'Confidential': self.CONFIDENTIAL,
        }
        if access_level in options:
            return options[access_level]
        else:
            raise MikeCloudException("Invalid project access level", access_level)


class ProjectMemberRole(Enum):
    READER = 100
    CONTRIBUTOR = 200
    OWNER = 300

    @classmethod
    def from_string(self, role: str):
        options = {
            'Reader': self.READER,
            'Contributor': self.CONTRIBUTOR,
            'Owner': self.OWNER,
        }
        if role in options:
            return options[role]
        else:
            raise MikeCloudException("Invalid project member role", role)


class ProjectMemberInput:
    def __init__(self, user_id: str, role: ProjectMemberRole) -> None:
        self._data = {
            "userId": user_id,
            "role": role.name.title()
        }
    
    def body(self):
        return self._data


class ProjectMemberOutput:
    def __init__(self, user_id : str, role: ProjectMemberRole):
        self._user_id = user_id
        self._role = role
    
    @property
    def user_id(self):
        self._user_id
    
    @property
    def role(self):
        self._role

    @classmethod
    def from_body(cls, data: dict):
        user_id = data["userId"]
        role = ProjectMemberRole.from_string(data["role"])
        return cls(user_id, role)


class CreateProjectInput:
    def __init__(self, name: str, description: str = None, access_level: AccessLevel = AccessLevel.SHARED, metadata: dict = {}, members: list = []):
        self._name = name
        self._access_level = access_level
        self._description = "" if description is None else description
        self._metadata = metadata
        self._settings = {}
        self._members = members

    @property 
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._description = description
    
    @property 
    def access_level(self):
        return self._access_level

    @access_level.setter
    def access_level(self, access_level):
        self._access_level = access_level

    def set_metadata_item(self, key, value):
        self._metadata[key] = value

    def body(self):
        return {
            "name": self._name,
            "accessLevel": self._access_level.name.title(),
            "description": self._description,
            "metadata": self._metadata,
            "settings": self._settings,
            "members": [m.body() for m in self._members]
        }


class ProjectOutput:
    def __init__(
        self,
        id: str,
        created_at: datetime.date,
        created_by: str,
        updated_at: datetime.date,
        updated_by: str,
        deleted_at: datetime.date,
        deleted_by: str,
        name: str,
        description: str,
        metadata: dict,
        settings: dict,
        access_level: AccessLevel,
        members: list,
        capabilities: list,
        parent_project_id: str, 
        inherits_members: bool,
        row_version: str
    ):
        self._id = id
        self._created_at = created_at
        self._created_by = created_by
        self._updated_at = updated_at
        self._updated_by = updated_by
        self._deleted_at = deleted_at
        self._deleted_by = deleted_by
        self._name = name
        self._description = description
        self._metadata = metadata
        self._settings = settings
        self._access_level = access_level
        self._members = members
        self._capabilities = capabilities
        self._parent_project_id = parent_project_id
        self._inherits_members = inherits_members
        self._row_version = row_version

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name
    
    @property
    def description(self):
        return self._description

    @property
    def created_at(self):
        return self._created_at
    
    @property
    def created_by(self):
        return self._created_by

    @property
    def updated_at(self):
        return self._updated_at

    @property
    def updated_by(self):
        return self._updated_by

    @property
    def deleted_by(self):
        return self._deleted_by

    @property
    def deleted_at(self):
        return self._deleted_at

    @property
    def metadata(self) -> dict:
        return self._metadata

    @property
    def settings(self) -> dict:
        return self._settings
    
    @property
    def access_level(self) -> AccessLevel:
        return self._access_level
    
    @property
    def capabilities(self) -> dict:
        return self._capabilities

    @property
    def parent_project_id(self) -> str:
        return self._parent_project_id

    @property
    def inherits_members(self) -> bool:
        return self._inherits_members
    
    @property
    def row_version(self) -> str:
        return self._row_version
    
    @property
    def capabilities(self) -> dict:
        return self._capabilities

    @property
    def members(self) -> list:
        return self._members

    @classmethod
    def from_body(cls, body: dict):
        updated_at = None if body.get("updatedAt", None) is None else datetime.datetime.strptime(body["updatedAt"][:26], constants.DATETIMEFORMAT)
        deleted_at = None if body.get("deletedAt", None) is None else datetime.datetime.strptime(body["deletedAt"][:26], constants.DATETIMEFORMAT)
        return cls(
            id = body["id"],
            created_at = datetime.datetime.strptime(body["createdAt"][:26], constants.DATETIMEFORMAT),
            created_by = body["createdBy"],
            updated_at = updated_at,
            updated_by = body["updatedBy"],
            deleted_at = deleted_at,
            deleted_by = None if "deletedBy" not in body else body["deletedBy"],
            name = body["name"],
            description = None if "description" not in body else body["description"],
            metadata = body["metadata"],
            settings = body["settings"],
            access_level = AccessLevel.from_string(body["accessLevel"]),
            members = [ProjectMemberOutput.from_body(m) for m in body["members"]],
            capabilities = body["capabilities"],
            parent_project_id = None if "parentProjectId" not in body else body["parentProjectId"],
            inherits_members = body["inheritsMembers"],
            row_version = body["rowVersion"]
        )


class ListProjectOutput:
    def __init__(
        self,
        id: str,
        created_at: datetime.date,
        created_by: str,
        updated_at: datetime.date,
        updated_by: str,
        deleted_at: datetime.date,
        deleted_by: str,
        name: str,
        description: str,
        access_level: AccessLevel,
        parent_project_id: str, 
        row_version: str
    ):
        self._id = id
        self._created_at = created_at
        self._created_by = created_by
        self._updated_at = updated_at
        self._updated_by = updated_by
        self._deleted_at = deleted_at
        self._deleted_by = deleted_by
        self._name = name
        self._description = description
        self._access_level = access_level
        self._parent_project_id = parent_project_id
        self._row_version = row_version

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name
    
    @property
    def description(self):
        return self._description

    @property
    def created_at(self):
        return self._created_at
    
    @property
    def created_by(self):
        return self._created_by

    @property
    def updated_at(self):
        return self._updated_at

    @property
    def updated_by(self):
        return self._updated_by

    @property
    def deleted_by(self):
        return self._deleted_by

    @property
    def deleted_at(self):
        return self._deleted_at
    
    @property
    def access_level(self) -> AccessLevel:
        return self._access_level

    @property
    def parent_project_id(self) -> str:
        return self._parent_project_id
    
    @property
    def row_version(self) -> str:
        return self._row_version

    @classmethod
    def from_body(cls, body: dict):
        updated_at = None if "updatedAt" not in body else datetime.datetime.strptime(body["updatedAt"][:26], constants.DATETIMEFORMAT)
        deleted_at = None if "DeletedAt" not in body else datetime.datetime.strptime(body["deletedAt"][:26], constants.DATETIMEFORMAT)
        return cls(
            id = body["id"],
            created_at = datetime.datetime.strptime(body["createdAt"][:26], constants.DATETIMEFORMAT),
            created_by = body["createdBy"],
            updated_at = updated_at,
            updated_by = body["updatedBy"],
            deleted_at = deleted_at,
            deleted_by = None if "deletedBy" not in body else body["deletedBy"],
            name = body["name"],
            description = None if "description" not in body else body["description"],
            access_level = AccessLevel.from_string(body["accessLevel"]),
            parent_project_id = None if "parentProjectId" not in body else body["parentProjectId"],
            row_version = body["rowVersion"]
        )


class RowVersionOutput:
    def __init__(self, id: str, row_version: str):
        self._id = id
        self._row_version = row_version

    @property
    def id(self):
        return self._id
    
    @property
    def row_version(self) -> str:
        return self._row_version

    @classmethod
    def from_body(cls, body: dict):
        return cls(
            id = body["id"],
            row_version = body["rowVersion"]
        )


class UpdateProjectInput:
    def __init__(self, id, name: str, description: str = None, metadata: dict = {}, settings: dict = {}):
        self._id = id
        self._name = name
        self._description = description
        self._metadata = metadata
        self._settings = settings
    
    @property 
    def id(self):
        return self._id

    @property 
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name
    
    @property 
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        self._description = description
    
    @property 
    def access_level(self):
        return self._access_level
    
    def set_metadata_item(self, key, value):
        self._metadata[key] = value

    def set_settings_item(self, key, value):
        self._settings[key] = value

    def body(self):
        body = {
            "id": self._id,
            "name": self._name
        }

        if self._description:
            body["description"] = self._description
        
        if self._metadata:
            body["metadata"] = self._metadata
        
        if self._settings:
            body["settings"] = self._settings
        
        return body


class ProjectMember:
    def __init__(self, user_id, role: ProjectMemberRole):
        self._user_id = user_id
        self._role = role
    
    @property 
    def user_id(self):
        return self._user_id

    @property 
    def role(self):
        return self._role
    
    def body(self):
        return {
            "userId": self._user_id,
            "role": self._role.name.title()
        }

    @classmethod
    def from_body(cls, body: dict):
        return cls(
            user_id=body["userId"],
            role=ProjectMemberRole.from_string(body["role"])
        )

class PlatformUser:
    def __init__(self, user_id, display_name: str, email: str, is_admin: bool, customer_id: str, customer_name: str) -> None:
        self._user_id = user_id
        self._display_name = display_name
        self._email = email
        self._is_admin = is_admin
        self._customer_id = customer_id
        self._customer_name = customer_name
    
    @property
    def user_id(self):
        return self._user_id
    
    @property
    def display_name(self):
        return self._display_name

    @property
    def email(self):
        return self._email

    @property
    def is_admin(self):
        return self._is_admin

    @property
    def customer_id(self):
        return self._customer_id

    @property
    def customer_name(self):
        return self._customer_name

    @classmethod
    def from_body(cls, body: dict):
        return cls(
            user_id = body["userId"],
            display_name = body["displayName"],
            email = body.get("email", None),
            is_admin = body.get("isAdmin", False),
            customer_id = body.get("customerId", None),
            customer_name = body.get("customerName", None)
        )

class ProjectCapabilities:
    def __init__(self,
        can_edit: bool,
        can_edit_access_level: bool,
        can_delete: bool,
        can_grant_access: bool,
        can_create_content: bool,
        can_list_content: bool,
        can_update_content: bool,
        can_delete_content: bool,
        can_read_content: bool
    ) -> None:
        self._can_edit = can_edit
        self._can_edit_access_level = can_edit_access_level,
        self._can_delete = can_delete,
        self._can_grant_access = can_grant_access,
        self._can_create_content = can_create_content,
        self._can_list_content = can_list_content,
        self._can_update_content = can_update_content,
        self._can_delete_content = can_delete_content,
        self._can_read_content = can_read_content

    @property
    def can_edit(self):
        return self._can_edit

    @property
    def can_edit_access_level(self):
        return self._can_edit_access_level

    @property
    def can_delete(self):
        return self.can_delete

    @property
    def can_grant_access(self):
        return self._can_grant_access

    @property
    def can_create_content(self):
        return self._can_create_content

    @property
    def can_list_content(self):
        return self._can_list_content

    @property
    def can_update_content(self):
        return self._can_update_content

    @property
    def can_delete_content(self):
        return self._can_delete_content

    @property
    def can_read_content(self):
        return self._can_read_content

    @classmethod
    def from_body(cls, body: dict):
        return cls(
            can_edit = body.get("canEdit", False),
            can_edit_access_level = body.get("canEditAccessLevel", False),
            can_delete = body.get("canDelete", False),
            can_grant_access = body.get("canGrantAccess", False),
            can_create_content = body.get("canCreateContent", False),
            can_list_content = body.get("canListContent", False),
            can_update_content = body.get("canUpdateContent", False),
            can_delete_content = body.get("canDeleteContent", False),
            can_read_content = body.get("canReadContent", False)
        )

class ProjectPathNode:
    def __init__(
        self,
        id,
        name,
        parent_project_id,
        is_deleted,
        capabilities,
        access_level,
        inherit_members,
        effective_user_role
    ) -> None:
        self._id = id
        self._name = name
        self._parent_project_id = parent_project_id
        self._is_deleted = is_deleted
        self._capabilities = capabilities
        self._access_level = access_level
        self._inherit_members = inherit_members
        self._effective_user_role = effective_user_role

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def parent_project_id(self):
        return self._parent_project_id

    @property
    def is_deleted(self):
        return self._is_deleted

    @property
    def capabilities(self):
        return self._capabilities

    @property
    def access_level(self):
        return self._access_level

    @property
    def inherit_members(self):
        return self._inherit_members

    @property
    def effective_user_role(self):
        return self._effective_user_role


    @classmethod
    def from_body(cls, body: dict):
        return cls(
            id = body["id"],
            name = body["name"],
            parent_project_id = body.get("parentProjectId", None),
            is_deleted = body.get("isDeleted", True),
            capabilities = ProjectCapabilities.from_body(body["capabilities"]),
            access_level = AccessLevel.from_string(body.get("accessLevel", AccessLevel.SHARED.name.title())),
            inherit_members = body.get("inheritsMembers", True),
            effective_user_role = ProjectMemberRole.from_string(body.get("effectiveUserRole", ProjectMemberRole.READER.name.title()))
        )
        
class SubprojectInput:
    def __init__(
        self,
        name: str,
        access_level: AccessLevel = None,
        description: str = None,
        metadata: dict = None,
        settings: dict = None,
        members: List[ProjectMember] = None
    ) -> None:
        self._name = name
        self._access_level = access_level
        self._description = description
        self._metadata = metadata
        self._settings = settings
        self._members = members

    def body(self) -> dict:
        body = {
            "name": self._name
        }

        if self._access_level:
            body["accessLevel"] = self._access_level

        if self._description:
            body["description"] = self._description
        
        if self._metadata:
            body["metadata"] = self._metadata
        
        if self._settings:
            body["settings"] = self._settings
        
        if self._members:
            body["members"] = [m.body() for m in self._members]
        
        return body

class PrepareHierarchyInput(PrepareHierarchyInputV3):
    pass

class PrepareHierarchyOutput(PrepareHierarchyOutputV3):
    pass

class PathAction(PathActionV3):
    
    @classmethod
    def create(cls, path:str, is_folder:bool=False):
        return PathActionCreateV3(path, is_folder)
    
    @classmethod
    def create_if_not_exists(cls, path:str, is_folder:bool=False):
        return PathActionCreateIfNotExistsV3(path, is_folder)
    
    @classmethod
    def delete(cls, path:str, is_folder:bool=False):
        return PathActionDeleteV3(path, is_folder)


class DatasetOutput(DatasetOutputV3):
    pass


class DatasetRecursiveListOutput(DatasetRecursiveListOutputV3):
    pass

    
UnitId = UnitIdV3
ItemId = ItemIdV3

class MetadataClientV1(MetadataGenClientV1):
    def __init__(self, inspectFnc=MetadataGenClientV1.DefaultInspectFnc, **kwargs):
        super().__init__(inspectFnc, **kwargs)


class MetadataClientV2(MetadataGenClientV2):
    def __init__(self, inspectFnc=MetadataGenClientV2.DefaultInspectFnc, **kwargs):
        super().__init__(inspectFnc, **kwargs)


class MetadataClientV3(MetadataGenClientV3):
    def __init__(self, inspectFnc=MetadataGenClientV3.DefaultInspectFnc, **kwargs):
        super().__init__(inspectFnc, **kwargs)


class MetadataClient():
    
    def __init__(self, inspectFnc=MetadataGenClientV1.DefaultInspectFnc, **kwargs):
        self._client1 = MetadataClientV1(inspectFnc, **kwargs)
        self._client2 = MetadataClientV2(inspectFnc, **kwargs)
        self._client3 = MetadataClientV3(inspectFnc, **kwargs)

    def list_projects(self, nameprefix=None) -> Generator[ListProjectOutput, None, None]:
        is_first_query = True
        cursor = None
        while cursor or is_first_query:
            response = self._client3.GetProjectListV3(nameprefix, role=None, capability=None, sortby=None, sortorder=None, cursor=cursor, limit=1000)
            is_first_query = False
            cursor = response.Body.get("cursor", None)
            for p in response.Body['data']:
                yield ListProjectOutput.from_body(p)

    def get_project(self, project_id) -> ProjectOutput:
        response = self._client3.GetProjectV3(project_id)
        return ProjectOutput.from_body(response.Body)

    def create_project(self, project_input: CreateProjectInput) -> ProjectOutput:
        response = self._client3.CreateProjectV3(project_input.body())
        return ProjectOutput.from_body(response.Body)

    def delete_project(self, project_id, permanently=False) -> bool:
        response = self._client3.DeleteProjectV3(project_id)
        if permanently:
            response = self._client2.DestroyProjectV2(project_id)
        return response.IsOk            
    
    def update_project(self, update_input: UpdateProjectInput) -> ProjectOutput:
        """
        Update selected project properties.

        This is a hard update in the sense that you must include all properties in the 'update_intput' parameter.
        If you do not include a property in the 'update_intput' parameter, or you set it to None, that property will be set to None by the update.
        For 'metadata' and 'settings' parts, if key is absent in the input, the key and its value will be dropped by the update.
        If it is set to None, the value for that key will be set to None by the update.
        """
        project = self.get_project(update_input.id)
        body = update_input.body()
        body["rowVersion"] = project.row_version
        response = self._client3.UpdateProjectV3(body)
        return ProjectOutput.from_body(response.Body)

    def update_project_access_level(self, project_id, access_level: AccessLevel) -> RowVersionOutput:
        project = self.get_project(project_id)
        body = {
            "id": project_id,
            "accessLevel": access_level.name.title(),
            "rowVersion": project.row_version
        }
        response = self._client3.UpdateProjectAccessLevelV3(body)
        return RowVersionOutput.from_body(response.Body)
    
    def get_project_members(self, project_id) -> List[ProjectMember]:
        response = self._client3.GetProjectMemberV3(project_id)
        body = response.Body
        return [ProjectMember.from_body(d) for d in body["data"]]

    def add_project_member(self, project_id, member: ProjectMember) -> ProjectMember:
        """Add one project member to a project"""
        response = self._client3.CreateProjectMemberV3(member.body(), project_id)
        return ProjectMember.from_body(response.Body)

    def set_project_members(self, project_id, members: List[ProjectMember]) -> RowVersionOutput:
        """Replace all existing project members with new 'members'"""
        project = self.get_project(project_id)
        body = {
            "members": [m.body() for m in members],
            "rowVersion": project.row_version
        }
        response = self._client3.SetProjectMembersV3(body, project_id)
        return RowVersionOutput.from_body(response.Body)
    
    def remove_project_member(self, project_id, member_id) -> bool:
        response = self._client3.DeleteProjectMemberV3(member_id, project_id)
        response.IsOk
    
    def inherit_project_members(self, project_id) -> RowVersionOutput:
        """Set project to inherit all members from its parent project"""
        project = self.get_project(project_id)
        body = { "rowVersion": project.row_version }
        response = self._client3.SetInheritMembersV3(body, project_id)
        return RowVersionOutput.from_body(response.Body)

    def get_users(self) -> List[PlatformUser]:
        response = self._client3.GetUserList()
        return [PlatformUser.from_body(u) for u in response.Body["data"]]

    def get_project_capabilities(self, project_id) -> ProjectCapabilities:
        response = self._client3.GetProjectCapabilitiesV3(project_id)
        return ProjectCapabilities.from_body(response.Body)
    
    def move_project(self, source_project_id, target_project_id) -> bool:
        body = {"targetProjectId": target_project_id}
        response = self._client3.MoveProjectV3(body, source_project_id)
        return response.IsOk

    def prepare_hierarchy(self, project_id, actions:Tuple[PathAction], default_access_level:AccessLevelV3=AccessLevelV3.SHARED, sas_token_expiration:datetime.timedelta=datetime.timedelta(hours=1)) -> PrepareHierarchyOutput:
        """Create hierarchy of subprojects and blob files in a project

        This method is intended for syncing folders between local system and the cloud platform.
        
        :param project_id: root project of the hierarchy.
        :param actions: Definition of the hierarchy as in iterable of PathAction instances like:
            [
                PathAction.create_if_not_exists("foo", True),
                PathAction.create("foo/bar"),
                PathAction.create("foo/spam", True),
                PathAction.create("foo/spam/eggs"),
                PathAction.delete("foo/fred", True)
            ]
        :param default_access_level: access level used for created subprojects if not specified explicitely
        :sas_token_expiration: Time for which the resulting sas tokens should be valid so that files can be uploaded. Default is 1 hour.
        :return: Object with iterable results property of path action outputs
        :rtype: PrepareHierarchyOutput
        """
        input = PrepareHierarchyInput(actions, defaultAccessLevel=default_access_level, sasTokenExpiration=str(sas_token_expiration))
        response = self._client3.PrepareHierarchy(input.to_dict(), project_id)
        return PrepareHierarchyOutput.from_dict(response.Body)

    def get_project_path(self, project_id) -> List[ProjectPathNode]:
        response = self._client3.GetProjectPathV3(project_id)
        data = response.Body["data"]
        return [ProjectPathNode.from_body(node) for node in data]
    
    def create_subproject(self, project_id, input: SubprojectInput) -> ProjectOutput:
        response = self._client3.CreateSubProjectV3(input.body(), project_id)
        return ProjectOutput.from_body(response.Body)

    def list_subprojects(self, project_id, nameprefix: str = None, role:ProjectMemberRole = None, sortby:str=None, descending=False) -> Generator[ListProjectOutput, None, None]:
        """
        List of direct subprojects of project 'project_id'.

        :param project_id: ID of the project to start from
        :param nameprefix: Search only for subproject that start with nameprefix
        :param role: return results where the effective user role the value of this parameter.
        :para sortby: can be one of Name, CreatedAt, UpdatedAt, AccessLevel
        :param descending: sort results in descending order if True, otherwise sort in ascending orded, default is False
        """
        limit = 1000
        is_first_query = True
        cursor = None
        sortorder = "Desc" if descending else "Asc"
        while cursor or is_first_query:
            response = self._client3.GetSubProjectListV3(project_id, nameprefix, role, capability=None, sortby=sortby, sortorder=sortorder, cursor=cursor, limit=limit)
            is_first_query = False
            cursor = response.Body.get("cursor", None)
            for p in response.Body['data']:
                yield ListProjectOutput.from_body(p)
        limit = 1000

    def list_datasets(self, project_id, includesastokens:bool=False) -> Generator[DatasetOutput, None, None]:
        """
        List datasets in a project
        
        :param project_id: ID of the project where the datasets reside
        :param includesastokens: if True, the response will include sas tokens where possible, default is False
        :return: Generator of DatasetOutput instances
        """
        response = self._client3.GetDatasetListV3(project_id, includesastokens=includesastokens)
        data = response.Body["data"]
        return (DatasetOutput.from_dict(d) for d in data)

    def list_datasets_recursive(self, project_id, includesastokens:bool=False, dataset_type:DatasetType=None) -> Generator[DatasetOutput, None, None]:
        """
        List datasets in a project recursively
        
        :param project_id: ID of the project where the datasets reside
        :param includesastokens: if True, the response will include sas tokens where possible, default is False
        :param dataset_type: Return only datasets of selected dataset type, default is None for all dataset types
        :return: Generator of DatasetRecursiveListOutput instances
        """
        limit = 1000
        offset = 0
        first_query = True
        while True:
            if not first_query:
                if not response.Body["data"]:
                    break
            response = self._client3.GetRecursiveDatasetListV3(project_id, offset=offset, limit=limit, includesastokens=includesastokens, datasettype=dataset_type)
            for i in response.Body["data"]:
                yield DatasetRecursiveListOutput.from_dict(i)
            offset = offset + limit + 1
            first_query = False

    def get_dataset(self, dataset_id):
        """
        Get dataset
        
        :param dataset_id: ID of the dataset
        :return: DatasetOutput
        :rtype: DatasetOutput
        """
        response = self._client3.GetDatasetV3(dataset_id)
        return DatasetOutput.from_dict(response.Body)

    def delete_dataset(self, dataset_id, permanently:bool=False) -> bool:
        """
        Delete dataset

        :param dataset_id: ID of the dataset
        :return: True if the dataset was deleted successfully
        :rtype: bool
        """
        response = self._client3.DeleteDatasetV3(dataset_id)
        if permanently:
            response = self._client2.DestroyDatasetV2(dataset_id)
        return response.IsOk

    def update_dataset(self, project_id, dataset_id, name:str=None, description:str=None, metadata:dict=None) -> DatasetOutput:
        """
        Update dataset

        :param project_id: ID of the project where the dataset resides
        :param dataset_id: ID of the dataset
        :param name: new name to use, no update to name is made if None, which is the default
        :param description: new description to use, no update to description is made if None, which is the default
        :param metadata: new metadata to use, no update to metadata is made if None, which is the default. If not empty, metadata will be completely replaced by the new entry.
        :return: DatasetOutput
        :rtype: DatasetOutput
        """
        dataset = self.get_dataset(dataset_id)
        datadict = dataset.to_dict()
        
        if name:
            datadict["name"] = name
        if description is not None:
            datadict["description"] = description
        if metadata is not None:
            datadict["metadata"] = metadata

        response = self._client3.UpdateDatasetV3(datadict, project_id)
        return DatasetOutput.from_dict(response.Body)

    def move_dataset(self, dataset_id, target_project_id) -> bool:
        """
        Move dataset to another project

        :param dataset_id: ID of the dataset
        :param target_project_id: ID of the project the dataset should be moved to
        :return: True if the dataset was moved successfully
        :rtype: bool
        """
        body = {"targetProjectId": target_project_id}
        response = self._client3.MoveDatasetV3(body, dataset_id)
        return response.IsOk

    def get_service_url(self, servicename) -> str:
        """Get service url by service name
        :param servicename: Short name of the service to get full URL for
        :return: Service URL
        :rtype: str
        """
        response = self._client3.GetServiceUrlV3(servicename)
        return response.Body["data"]

    def get_sas_token_string(self, project_id, resourse_id=None, expiration:datetime.timedelta=datetime.timedelta(hours=1)) -> str:
        """Get Shared Access Signature (SAS) token as a string
        :param project_id: ID of the projec to get SAS token for
        :param resource_id: Optional resource id to get SAS token for. This is typically the dataset ID.
        :param expiration: Time for which the token should be valid, default is 1 hour.
        :return: SAS token
        :rtype: str
        """
        response = self._client2.GetSasTokenV2(project_id, resourse_id, str(expiration))
        sas_token = response.Body["data"]
        return sas_token
    
    def wait_until_dataset_exists(self, dataset_id, timeout:float=7.0, timeout_handler=lambda: (_ for _ in ()).throw(MikeCloudException(f"Dataset did not exists within a specified timeout."))):
        """Wait until dataset exists in metadata.

        This is a utility method used for example when creating a time series service.
        The system has to wait until the dataset exists in metadata before it can say it exists.

        :param dataset_id: Dataset id to wait for.
        :param timeout: Optional. How many seconds to wait before a timeout handler is invoked.
        :param timeout_handler: Optional callable to call when timeout is reached.
        """
        elapsed_time = 0.0
        sleep_interval = 0.75
        dataset = None
        while dataset is None:
            try:
                dataset = self.get_dataset(dataset_id)
            except MikeCloudRestApiException as ex:
                if ex.status_code == 404 or ex.status_code == 403:
                    time.sleep(sleep_interval)
                    elapsed_time += sleep_interval
                    if(elapsed_time >= timeout):
                        timeout_handler()


if __name__ == '__main__':
    print(__file__)
    print(dir())



#!/usr/bin/env python

# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import argparse, datetime, fileinput, glob, heapq, json, os, os.path, re, random, requests, sys, urllib.request
from http.client import REQUEST_HEADER_FIELDS_TOO_LARGE, HTTPConnection, HTTPSConnection

def __initParser(parser):
    parser.add_argument("files", help="OpenAPI file", metavar="url", nargs="*")
    parser.add_argument("-c", "--classname", help="Class name", default="OpenapiClient")
    parser.add_argument("-n", "--classnum", help="First class number", type=int, default=1)
    parser.add_argument("-s", "--dhiservice", help="Platform service id")
    parser.add_argument("-r", "--headerparams", help="Header parameters", metavar="PARAMETERS")
    parser.add_argument("-i", "--inputtoheader", help="Input to header parameters", metavar="INPUTPARAMETERS")
    parser.add_argument("-k", "--kwargsfor", help="Operation to use **kwargs", metavar="OPERATIONS")
    parser.add_argument("-f", "--headersfor", help="Additional headers for endpoints", metavar="HEADERDEF")
    parser.add_argument("-t", "--datafields", help="Additional data fields for data contracts", metavar="FIELDDEF")
    parser.add_argument("-d", "--includedeprecated", help="Include deprecated methods", metavar="OPERATIONS")
    parser.add_argument("-o", "--output", default="-", help="Write output to a file")
    #parser.add_argument("--relative", help="show relative paths", action="store_false")
    #parser.add_argument("--transitivedeps", help="show transitive dependencies", action="store_false")
    #parser.add_argument("--shownotfound", help="show not found projects", action="store_true")
    #parser.add_argument("--shownotfounddeps", help="show not found dependencies", action="store_true")
    #parser.add_argument("-f", "--format", help="output format", choices=["plain", "json", "xml"], default="plain")
    #parser.add_argument("--debug", help="show debug output", action="store_true")
    parser.add_argument("-v", "--verbose", default=0, help="Make the operation more talkative (increase verbosity), show debug output", action="count")

def main():
    args = __parseBasic(description="Create Python wrapper classes from OpenAPI", init=__initParser)
    if args.verbose:
        print(f"args:{args}")

    if args.output and args.output != "-":
        with open(args.output, "w") as outfile:
            writeFiles(args, outfile)
    else:
        writeFiles(args, sys.stdout)
    #with urllib.urlopen(f) as response:
    #    obj = json.load(f)
    #    print(obj)
    #    #html = response.read()

def writeFiles(args, outfile):
    n = 0
    kwargsoperations = set()
    if args.kwargsfor:
        kwargsoperations.update(args.kwargsfor.split(","))
    headersfor = Helper.initheadersfor(args.headersfor)
    datafields = Helper.initdatafields(args.datafields)
    includedeprecated = set()
    if args.includedeprecated:
        includedeprecated.update(args.includedeprecated.split(","))
    headerparams = []
    if args.headerparams:
        headerparams.extend(args.headerparams.split(","))
    inputtoheader = {}
    if args.inputtoheader:
        inputtoheader.update(dict([reversed(x.split('=')) for x in args.inputtoheader.split(",")]))
    Helper.writeFileHeader(outfile)
    for inputfile in args.files:
        openapiObj = Helper.getOpenApiObj(inputfile)
        classname = f"{args.classname}{n+args.classnum}"
        suffix = f"V{n+args.classnum}"
        Helper.generateApiClasses(openapiObj, suffix, inputfile, classname, kwargsoperations, headersfor, datafields, headerparams, inputtoheader, includedeprecated, args.dhiservice, outfile)
        n += 1
        #with urllib.urlopen(f) as response:
        #    obj = json.load(f)
        #    print(obj)
        #    #html = response.read()

def __parseBasic(prog=None, description=None, init=None):
    parser = argparse.ArgumentParser(prog=prog, description=description)
    if init:
        init(parser)
    try:
        return parser.parse_args()
    except:
        sys.exit(1)

class ContractInfo:
    def __init__(self, name, content, dependson = set()):
        self.refcount = 0
        self.name = name
        self.content = content
        self.dependson = dependson
        self.dependsonme = set()
    def __lt__(self, other):
        return self.refcount < other.refcount

class Helper:
    TITLE = "title"
    DESCRIPTION = "description"
    VERSION = "version"
    NAME = "name"
    IN = "in"
    OPERATIONID = "operationId"
    SUMMARY = "summary"
    SCHEMA = "schema"
    DEFAULT = "default"
    REQUESTBODY = "requestBody"
    MethodNumbers = {}

    @classmethod
    def writeFileHeader(cls, out=sys.stdout):
        #print("#!/usr/bin/env python", file=out)
        print(f"# Generated using {os.path.basename(__file__)}", file=out)
        cmdline = " ".join([os.path.basename(x) if i == 0 else f"\"{x}\"" for i, x in enumerate(sys.argv)])
        print(f"# {cmdline}", file=out)
        print(f"# {datetime.datetime.utcnow()}Z", file=out)
        print("from typing import Any, Dict, List, NewType, Type, TypeVar, Union", file=out)
        print("from enum import Enum", file=out)
        print("import attr", file=out)
        print("from ..base.client import DataContract, PlatformClient, Response", file=out)

    @classmethod
    def getOpenApiObj(cls, inputfile):
        return requests.get(inputfile).json()

    @classmethod
    def initheadersfor(cls, headersfor):
        result = dict()
        if headersfor:
            for h in headersfor.split(","):
                parts = h.split(":")
                if len(parts) != 3:
                    sys.exit(f"wrong format in header info (endpoint:header:value): {h}")
                if op := parts[0]:
                    content = { parts[1]: f"\"{parts[2]}\"" }
                    if headers := result.get(op):
                        headers.update(content)
                    else:
                        result[op] = content
        return result

    @classmethod
    def initdatafields(cls, datafields):
        result = dict()
        if datafields:
            for d in datafields.split(","):
                parts = d.split(":")
                if len(parts) != 4:
                    sys.exit(f"wrong format in data field info (datacontract:field:type:value): {d}")
                if dc := parts[0]:
                    content = { parts[1]: f"{parts[2]} = \"{parts[3]}\"" }
                    if fields := result.get(dc):
                        fields.update(content)
                    else:
                        result[dc] = content
        return result

    @classmethod
    def generateApiClasses(cls, openapiObj, suffix, inputfile, className, kwargsoperations, headersfor, datafields, headerparams, inputtoheader, includedeprecated, service=None, out=sys.stdout):
        info = openapiObj["info"]
        print(file=out)
        print(file=out)
        cls.__writeApiInfo(info, inputfile, out)
        cls.__generateContracts(openapiObj, info, suffix, inputfile, className, kwargsoperations, datafields, headerparams, inputtoheader, includedeprecated, service, out)
        cls.__generateEndpointsClass(openapiObj, info, suffix, inputfile, className, kwargsoperations, headersfor, headerparams, inputtoheader, includedeprecated, service, out)

    @classmethod
    def __writeApiInfo(cls, info, inputfile, out):
        print(f"# {inputfile}", file=out)
        print(f"# {info[cls.TITLE]}", file=out)
        print(f"# {info[cls.DESCRIPTION]}", file=out)
        print(f"# {info[cls.VERSION]}", file=out)
    @classmethod
    def __generateContracts(cls, openapiObj, info, suffix, inputfile, className, kwargsoperations, datafields, headerparams, inputtoheader, includedeprecated, service, out):
        if components := openapiObj["components"]:
            if schemas := components["schemas"]:
                allcontracts = dict()
                for name, content in schemas.items():
                    allcontracts[name] = ContractInfo(name, content, cls.__getContractDependson(name, content))
                for name, info in allcontracts.items():
                    for dep in info.dependson:
                        if d := allcontracts.get(dep):
                            d.dependsonme.add(name)
                            info.refcount += 1
                contractslist = [x for x in allcontracts.values()]
                heapq.heapify(contractslist)
                while contractslist:
                    heapq.heapify(contractslist)
                    top = heapq.heappop(contractslist)
                    name = top.name
                    content = top.content
                    for dep in top.dependsonme:
                        if deponmeitem := allcontracts.get(dep):
                            deponmeitem.refcount -= 1
                    cls.__generateContract(name, datafields.get(name), content, openapiObj, info, suffix, inputfile, className, kwargsoperations, headerparams, inputtoheader, includedeprecated, service, out)
    @classmethod
    def __getContractDependson(cls, name, schema):
        result = set()
        if schema.get("type") == "object":
            if parentclassname := cls.__getParentClassType(schema, ""):
                result.add(parentclassname)
            properties = schema.get("properties")
            if properties:
                for _, propertyinfo in properties.items():
                    if propertytype := cls.__getParentClassType(propertyinfo, ""):
                        result.add(propertytype)
                    else:
                        if tmp := propertyinfo.get("type"):
                            if tmp == "array":
                                #TODO: if it is multidimensional array, it should get nested item type
                                result.add(cls.__getArrayItemType(propertyinfo, ""))
        return result
    @classmethod
    def __generateContract(cls, name, customfields, schema, openapiObj, info, suffix, inputfile, className, kwargsoperations, headerparams, inputtoheader, includedeprecated, service, out):
        classname = f"{name}{suffix}"
        print(file=out)
        if enuminfo := schema.get("enum"):
            print(f"class {classname}(str, Enum):", file=out)
            description = schema.get(cls.DESCRIPTION)
            cls.__printLines2([description], "    \"\"\"", "    ", "    \"\"\"", out=out)
            # enum values
            for e in enuminfo:
                print(f"    {e.upper()} = \"{e}\"", file=out)
            # methods
            print("    def __str__(self) -> str:", file=out)
            print("        return str(self.value)", file=out)
        elif schema.get("type") == "object":
            if not (parentclassname := cls.__getParentClassType(schema, suffix)):
                parentclassname = "DataContract"
            #print(f"{classname}Type = NewType(\"{classname}\", {parentclassname})", file=out)
            print(f"{classname}Type = TypeVar(\"{classname}Type\", bound=\"{classname}\")", file=out)
            print(file=out)
            print("@attr.s(auto_attribs=True)", file=out)
            print(f"class {classname}({parentclassname}):", file=out)
            description = schema.get(cls.DESCRIPTION)
            cls.__printLines2([description], "    \"\"\"", "    ", "    \"\"\"", out=out)
            # custom fields
            if customfields:
                for cfname, cfdef in customfields.items():
                    print(f"    {cfname}: {cfdef}", file=out)
            # properties
            renamedproperties = {}
            properties = schema.get("properties")
            if properties:
                for originalpropertyname, propertyinfo in properties.items():
                    propertyname = cls.__safeParam(originalpropertyname, {})
                    propertytype = cls.__getPropertyType(propertyinfo, suffix)
                    if originalpropertyname != propertyname:
                        renamedproperties[originalpropertyname] = propertyname
                    print(f"    {propertyname}: {propertytype} = None", file=out)
            renamedlist = ", ".join([f"\"{o}\": \"{p}\"" for o, p in renamedproperties.items()])
            print(f"    __renamed = {{ {renamedlist} }}", file=out)
            #renamedlist = ", ".join([f"\"{p}\": \"{o}\"" for o, p in renamedproperties.items()])
            #print(f"    __renamedr = {{ {renamedlist} }}", file=out)
            # methods
            print("    def to_dict(self) -> Dict[str, Any]:", file=out)
            print("        return self.get_dictionary(self.get_renamed())", file=out)
            print("    def load_dict(self, src_dict: Dict[str, Any]) -> None:", file=out)
            print("        DataContract.load_from_directory(self, src_dict, self.get_renamed())", file=out)
            print("    @classmethod", file=out)
            print("    def get_renamed(cls) -> Dict[str, str]:", file=out)
            print(f"        result = {parentclassname}.get_renamed().copy()", file=out)
            print(f"        result.update(cls.__renamed)", file=out)
            print(f"        return result", file=out)
            print("    @classmethod", file=out)
            print(f"    def from_dict(cls: {classname}Type, src_dict: Dict[str, Any]) -> {classname}Type:", file=out)
            print(f"        obj = {classname}()", file=out)
            print(f"        obj.load_dict(src_dict)", file=out)
            print(f"        return obj", file=out)
    @classmethod
    def __getPropertyType(cls, propertyinfo, suffix):
        if propertyinfo:
            propertytype = cls.__getParentClassType(propertyinfo, suffix)
            if propertytype:
                return propertytype
            if type := propertyinfo.get("type"):
                if type == "string":
                    return "str"
                elif type == "integer":
                    return "int"
                elif type == "number":
                    return "float"
                elif type == "array":
                    return f"List[{cls.__getArrayItemType(propertyinfo, suffix)}]"
                return "str"
            return cls.__getReftype(propertyinfo.get("$ref"), suffix)
        return None
    @classmethod
    def __getArrayItemType(cls, propertyinfo, suffix):
        if itemsinfo := propertyinfo.get("items"):
            if itemtype := cls.__getReftype(itemsinfo.get("$ref"), suffix):
                return itemtype
            return cls.__getPropertyType(itemsinfo, suffix)
        return None
    @classmethod
    def __getParentClassType(cls, schema, suffix):
        if schema:
            if allOf := schema.get("allOf"):
                for _, reftype in allOf[0].items():
                    return cls.__getReftype(reftype, suffix)
        return None
    @classmethod
    def __getReftype(cls, reftype, suffix):
        if reftype:
            if reftype == "string":
                return "str"
            elif reftype == "integer":
                return "int"
            elif reftype == "number":
                return "float"
            elif (reftype.startswith("#/components/schemas/")):
                tmp = reftype.replace("#/components/schemas/", "")
                return f"{tmp}{suffix}"
        return reftype
    @classmethod
    def __generateEndpointsClass(cls, openapiObj, info, suffix, inputfile, className, kwargsoperations, headersfor, headerparams, inputtoheader, includedeprecated, service, out):
        print(file=out)
        print(f"class {className}(PlatformClient):", file=out)
        print(f"    def __init__(self, inspectFnc=PlatformClient.DefaultInspectFnc, **kwargs):", file=out)
        includeheaders = f", includeheaders=PlatformClient.GetServiceHeaders({cls.__getInitParam(service)})" if service else ""
        print(f"        super().__init__(inspectFnc{includeheaders}, **kwargs)", file=out)
        headerparamsstr1 = ", ".join(headerparams)
        headerparamsstr1 = f", {headerparamsstr1}" if headerparamsstr1 else headerparamsstr1
        headerparamsstr2 = ", ".join([f"{x}={x}" for x in headerparams]) if headerparams else ""
        headerparamsstr2 = f", {headerparamsstr2}" if headerparamsstr2 else headerparamsstr2
        for url, ops in openapiObj["paths"].items():
            for method, opObj in ops.items():
                operation = cls.__getOperationId(opObj, method)
                if operation and (operation in includedeprecated or not opObj.get("deprecated")):
                    renamedparams = {}
                    queryparams = []
                    queryoptparams = []
                    pathparams = []
                    headers = {}
                    headerparams = {}
                    methodwithbody = method in ["post", "put", "patch"]
                    bodyparam = ", body" if methodwithbody and cls.REQUESTBODY in opObj else ""
                    callbodyparam = bodyparam if bodyparam else (", None" if methodwithbody else "")
                    if parameterslist := opObj.get("parameters"):
                        for p in parameterslist:
                            paramname = p[cls.NAME]
                            renamedparams[paramname] = cls.__safeParam(paramname.lower(), renamedparams)

                            if tmpH := inputtoheader.get(paramname):
                                headerparams[tmpH] = renamedparams[paramname]
                            if (tmpIn := p[cls.IN]) == "query":
                                pschema = p[cls.SCHEMA]
                                if pschema.get("nullable") or pschema.get("default") is None:
                                    queryoptparams.append(paramname)
                                else:
                                    queryparams.append(paramname)
                            elif tmpIn == "path":
                                pathparams.append(paramname)
                            elif tmpIn == "header":
                                headers[paramname] = p[cls.SCHEMA].get(cls.DEFAULT)
                    print(file=out)
                    hfparams = []
                    if hf := headersfor.get(operation):
                        hfparams = [f"{h}={v}" for h, v in hf.items()]
                    hfparamsstr = ", ".join(hfparams)
                    hfparamsstr = f", {hfparamsstr}" if hfparamsstr else ""
                    kwargsparam = ", **kwargs" if operation in kwargsoperations or "*" in kwargsoperations else ""
                    print(f"    def {operation}(self{headerparamsstr1}{bodyparam}{cls.__convertParams(pathparams, renamedparams)}{cls.__convertParams(queryparams, renamedparams)}{cls.__convertOptParams(queryoptparams, renamedparams)}{kwargsparam}) -> Response:", file=out)
                    doclines = []
                    if tmp := opObj.get(cls.SUMMARY):
                        doclines.append(tmp)
                    for tmp in opObj["tags"]:
                        doclines.append(tmp)
                    doclines.append(f"{method.upper()} {url}")
                    cls.__printLines2(doclines, "        \"\"\"", "        ", "        \"\"\"", out=out)
                    if queryparams or queryoptparams:
                        allqueryparams = queryparams+queryoptparams
                        reservedparams = [x for x in allqueryparams if cls.__isReserved(x)]
                        otherparams = [x for x in allqueryparams if not cls.__isReserved(x)]
                        if reservedparams:
                            print(f"        kw = {{{cls.__convertReservedParams(reservedparams, renamedparams)}}}", file=out)
                            kwparam = ", **kw" if otherparams else "**kw"
                        else:
                            kwparam = ""
                        print(f"        queryparams = self.GetQueryParams({cls.__convertQueryParams(otherparams, renamedparams)}{kwparam})", file=out)
                        qp = "queryparams"
                    else:
                        qp = "None"
                    ppf = "f" if pathparams else ""
                    requestUrl = cls.__convertUrl(url, renamedparams) if ppf else url
                    print(f"        return self.{method.capitalize()}Request({ppf}\"{requestUrl}\"{callbodyparam}, {qp}{cls.__convertHeaders(headers)}{headerparamsstr2}{cls.__convertHeaders(headerparams, params=True)}{hfparamsstr}{kwargsparam})", file=out)
    @staticmethod
    def __printLines(lines, prefix, suffix="", out=sys.stdout):
        if lines:
            for l in lines.splitlines():
                print(f"{prefix}{l}{suffix}", file=out)
    @staticmethod
    def __printLines2(lines, prefix1, prefix2, suffix="", out=sys.stdout):
        if lines:
            firstline = True
            for line in lines:
                if line:
                    isfirst = firstline
                    for l in line.splitlines():
                        if firstline:
                            print(f"{prefix1}{l}", file=out)
                            firstline = False
                        else:
                            print(f"{prefix2}{l}", file=out)
                    if isfirst and not firstline:
                        print("", file=out)
            if not firstline:
                print(suffix, file=out)
    @classmethod
    def __getOperationId(cls, op, method):
        #result = op.get(cls.OPERATIONID)
        #tmp = "operation"
        #return result if result else f"{method.capitalize()}Operation{cls.__nextNum(tmp)}"
        result = op.get(cls.OPERATIONID)
        return result.replace(" ", "_") if result else None
    @classmethod
    def __nextNum(cls, method):
        num = cls.MethodNumbers.get(method)
        num = num+1 if num else 1
        cls.MethodNumbers[method] = num
        return num
    @staticmethod
    def __isReserved(param):
        return param.find("@") >= 0 or param in ["from", "def", "import", "class", "return", "if", "else", "in", "body", "array"]
    @staticmethod
    def __mangleReserved(param, renamedparams):
        attempt = 0
        p = param
        while True:
            attempt += 1
            if attempt == 1 and p.find("@") >= 0:
                p = p.replace("@", "_")
            else:
                #p = f"{param}param{random.randint(0, 1000)}"
                p = f"{p}_"
            if p not in renamedparams:
                return p 
    @classmethod
    def __safeParam(cls, param, renamedparams):
        return cls.__mangleReserved(param, renamedparams) if cls.__isReserved(param) else param
    @classmethod
    def __convertReservedParams(cls, params, renamedparams):
        return ", ".join([f"\"{x}\": {renamedparams.get(x)}" for x in params])
    @classmethod
    def __convertParams(cls, params, renamedparams, sep=", ", withprefix=True):
        result = sep.join([renamedparams.get(x) for x in params])
        return f"{sep}{result}" if result and withprefix else result
    @classmethod
    def __convertOptParams(cls, params, renamedparams, sep=", ", withprefix=True):
        result = sep.join([f"{renamedparams.get(x)}=None" for x in params])
        return f"{sep}{result}" if result and withprefix else result
    @staticmethod
    def __convertQueryParams(params, renamedparams, sep=", ", withprefix=False):
        result = sep.join([f"{x}={renamedparams.get(x)}" for x in params])
        return f"{sep}{result}" if result and withprefix else result
    @staticmethod
    def __convertHeaders(headers, sep=", ", withprefix=True, params=False):
        fixvarname = lambda x: x.replace("-", "_")
        rightside = (lambda x: x) if params else (lambda x: f"\"{x}\"")
        tmp = [f"{fixvarname(n)}={rightside(v)}" for n, v in headers.items()]
        result = sep.join(tmp)
        return f"{sep}{result}" if result and withprefix else result
    @staticmethod
    def __convertUrl(url, renamedparams):
        pattern = re.compile(r'{([^}]*)}')
        return pattern.sub(lambda m: f"{{{renamedparams.get(m.group(1))}}}", url)
    @staticmethod
    def __getInitParam(value):
        return f"\"{value}\"" if value else "None"

if __name__ == '__main__':
    main()



# Generated using openapi2py.py
# openapi2py.py "--dhiservice" "engine" "--classname" "EngineGenClientV" "-n" "2" "-r" "projectid" "-f" "RunExecutionWithPlatformData:recursivetoken:true" "--out" "C:\work\devops\mike-platform-sdk-py\src\dhi\platform\generated\enginegen.py" "https://apispec-mike-platform-dev0.eu.mike-cloud-dev0.com/engine/v2"
# 2022-04-28 08:13:38.410516Z
from typing import Any, Dict, List, NewType, Type, TypeVar, Union
from enum import Enum
import attr
from ..base.client import DataContract, PlatformClient, Response


# https://apispec-mike-platform-dev0.eu.mike-cloud-dev0.com/engine/v2
# engine - Version 2
# API for executing modelling engines
# 2

GetConfigurationOutputV2Type = TypeVar("GetConfigurationOutputV2Type", bound="GetConfigurationOutputV2")

@attr.s(auto_attribs=True)
class GetConfigurationOutputV2(DataContract):
    """Characteristics of a pool type supported by engine/tool execution

    """
    poolType: str = None
    virtualMachineSize: str = None
    nodeLimit: int = None
    gpuCount: int = None
    numberOfCores: int = None
    memoryInMB: int = None
    resourceDiskSizeInMB: int = None
    description: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: GetConfigurationOutputV2Type, src_dict: Dict[str, Any]) -> GetConfigurationOutputV2Type:
        obj = GetConfigurationOutputV2()
        obj.load_dict(src_dict)
        return obj

EngineDefinitionV2Type = TypeVar("EngineDefinitionV2Type", bound="EngineDefinitionV2")

@attr.s(auto_attribs=True)
class EngineDefinitionV2(DataContract):
    name: str = None
    version: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineDefinitionV2Type, src_dict: Dict[str, Any]) -> EngineDefinitionV2Type:
        obj = EngineDefinitionV2()
        obj.load_dict(src_dict)
        return obj

EngineExecutionPeekParametersV2Type = TypeVar("EngineExecutionPeekParametersV2Type", bound="EngineExecutionPeekParametersV2")

@attr.s(auto_attribs=True)
class EngineExecutionPeekParametersV2(DataContract):
    """Describes the parameters used for peeking an execution

    """
    peekFileLocalPaths: List[str] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionPeekParametersV2Type, src_dict: Dict[str, Any]) -> EngineExecutionPeekParametersV2Type:
        obj = EngineExecutionPeekParametersV2()
        obj.load_dict(src_dict)
        return obj

GetConfigurationOutputCollectionResponseV2Type = TypeVar("GetConfigurationOutputCollectionResponseV2Type", bound="GetConfigurationOutputCollectionResponseV2")

@attr.s(auto_attribs=True)
class GetConfigurationOutputCollectionResponseV2(DataContract):
    data: List[GetConfigurationOutputV2] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: GetConfigurationOutputCollectionResponseV2Type, src_dict: Dict[str, Any]) -> GetConfigurationOutputCollectionResponseV2Type:
        obj = GetConfigurationOutputCollectionResponseV2()
        obj.load_dict(src_dict)
        return obj

EngineRunOutputV2Type = TypeVar("EngineRunOutputV2Type", bound="EngineRunOutputV2")

@attr.s(auto_attribs=True)
class EngineRunOutputV2(DataContract):
    """Describes the output when starting a new execution

    """
    executionId: str = None
    outputLocation: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineRunOutputV2Type, src_dict: Dict[str, Any]) -> EngineRunOutputV2Type:
        obj = EngineRunOutputV2()
        obj.load_dict(src_dict)
        return obj

EngineRunParameterV2Type = TypeVar("EngineRunParameterV2Type", bound="EngineRunParameterV2")

@attr.s(auto_attribs=True)
class EngineRunParameterV2(DataContract):
    """Optional, engine/tool-specific parameter, that can be used when starting on execution

    """
    name: str = None
    value: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineRunParameterV2Type, src_dict: Dict[str, Any]) -> EngineRunParameterV2Type:
        obj = EngineRunParameterV2()
        obj.load_dict(src_dict)
        return obj

EngineModelItemV2Type = TypeVar("EngineModelItemV2Type", bound="EngineModelItemV2")

@attr.s(auto_attribs=True)
class EngineModelItemV2(DataContract):
    """Describes an execution input from the MIKE Cloud Platform

    """
    subprojectId: str = None
    modelFileName: str = None
    engine: str = None
    version: str = None
    resultsRelativePath: str = None
    overwriteResultsIfExists: str = None
    runParameters: List[EngineRunParameterV2] = None
    reportLogUpdatesLines: int = None
    logFiles: List[str] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineModelItemV2Type, src_dict: Dict[str, Any]) -> EngineModelItemV2Type:
        obj = EngineModelItemV2()
        obj.load_dict(src_dict)
        return obj

class SortOrderV2(str, Enum):
    ASC = "Asc"
    DESC = "Desc"
    def __str__(self) -> str:
        return str(self.value)

ProblemDetailsV2Type = TypeVar("ProblemDetailsV2Type", bound="ProblemDetailsV2")

@attr.s(auto_attribs=True)
class ProblemDetailsV2(DataContract):
    type: str = None
    title: str = None
    status: int = None
    detail: str = None
    instance: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: ProblemDetailsV2Type, src_dict: Dict[str, Any]) -> ProblemDetailsV2Type:
        obj = ProblemDetailsV2()
        obj.load_dict(src_dict)
        return obj

ValidationProblemDetailsV2Type = TypeVar("ValidationProblemDetailsV2Type", bound="ValidationProblemDetailsV2")

@attr.s(auto_attribs=True)
class ValidationProblemDetailsV2(ProblemDetailsV2):
    errors: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = ProblemDetailsV2.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: ValidationProblemDetailsV2Type, src_dict: Dict[str, Any]) -> ValidationProblemDetailsV2Type:
        obj = ValidationProblemDetailsV2()
        obj.load_dict(src_dict)
        return obj

EngineGetOutputV2Type = TypeVar("EngineGetOutputV2Type", bound="EngineGetOutputV2")

@attr.s(auto_attribs=True)
class EngineGetOutputV2(DataContract):
    """Describes the details related to a supported engine/tool

    """
    name: str = None
    description: str = None
    allowedRunParameters: List[str] = None
    versions: List[str] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineGetOutputV2Type, src_dict: Dict[str, Any]) -> EngineGetOutputV2Type:
        obj = EngineGetOutputV2()
        obj.load_dict(src_dict)
        return obj

EngineGetOutputCollectionResponseV2Type = TypeVar("EngineGetOutputCollectionResponseV2Type", bound="EngineGetOutputCollectionResponseV2")

@attr.s(auto_attribs=True)
class EngineGetOutputCollectionResponseV2(DataContract):
    data: List[EngineGetOutputV2] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineGetOutputCollectionResponseV2Type, src_dict: Dict[str, Any]) -> EngineGetOutputCollectionResponseV2Type:
        obj = EngineGetOutputCollectionResponseV2()
        obj.load_dict(src_dict)
        return obj

EngineExecutionDiagnosticsOutputV2Type = TypeVar("EngineExecutionDiagnosticsOutputV2Type", bound="EngineExecutionDiagnosticsOutputV2")

@attr.s(auto_attribs=True)
class EngineExecutionDiagnosticsOutputV2(DataContract):
    """Describes the execution diagnostics

    """
    executionId: str = None
    diagnosticsLocation: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionDiagnosticsOutputV2Type, src_dict: Dict[str, Any]) -> EngineExecutionDiagnosticsOutputV2Type:
        obj = EngineExecutionDiagnosticsOutputV2()
        obj.load_dict(src_dict)
        return obj

OptionsV2Type = TypeVar("OptionsV2Type", bound="OptionsV2")

@attr.s(auto_attribs=True)
class OptionsV2(DataContract):
    """Setup of the hardware configuration used for the execution

    """
    poolType: str = None
    nodeCount: int = None
    maxExecutionElapsedTimeHours: float = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: OptionsV2Type, src_dict: Dict[str, Any]) -> OptionsV2Type:
        obj = OptionsV2()
        obj.load_dict(src_dict)
        return obj

EngineOutputItemV2Type = TypeVar("EngineOutputItemV2Type", bound="EngineOutputItemV2")

@attr.s(auto_attribs=True)
class EngineOutputItemV2(DataContract):
    """Describes the location in a blob storage where the result files for the execution are stored

    """
    uri: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineOutputItemV2Type, src_dict: Dict[str, Any]) -> EngineOutputItemV2Type:
        obj = EngineOutputItemV2()
        obj.load_dict(src_dict)
        return obj

EngineExecutionInputParametersV2Type = TypeVar("EngineExecutionInputParametersV2Type", bound="EngineExecutionInputParametersV2")

@attr.s(auto_attribs=True)
class EngineExecutionInputParametersV2(DataContract):
    """Describes the parameters of an execution that uses the MIKE Cloud Platform

    """
    models: List[EngineModelItemV2] = None
    output: EngineOutputItemV2 = None
    options: OptionsV2 = None
    scenarioName: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionInputParametersV2Type, src_dict: Dict[str, Any]) -> EngineExecutionInputParametersV2Type:
        obj = EngineExecutionInputParametersV2()
        obj.load_dict(src_dict)
        return obj

EngineOutputPlatformItemV2Type = TypeVar("EngineOutputPlatformItemV2Type", bound="EngineOutputPlatformItemV2")

@attr.s(auto_attribs=True)
class EngineOutputPlatformItemV2(DataContract):
    """Describes the location in the MIKE Platform Cloud where the result files for an execution associated with a setup file are stored

    """
    modelFileName: str = None
    resultsRelativePath: str = None
    overwriteResultsIfExists: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineOutputPlatformItemV2Type, src_dict: Dict[str, Any]) -> EngineOutputPlatformItemV2Type:
        obj = EngineOutputPlatformItemV2()
        obj.load_dict(src_dict)
        return obj

class EngineExecutionStatusV2(str, Enum):
    PENDING = "Pending"
    SETTINGUPCOMPUTERESOURCES = "SettingUpComputeResources"
    EVALUATINGINPUTSIZE = "EvaluatingInputSize"
    DOWNLOADINGINPUTFILES = "DownloadingInputFiles"
    INPROGRESS = "InProgress"
    UPLOADINGRESULTS = "UploadingResults"
    SUCCESS = "Success"
    FAILURE = "Failure"
    CANCELLING = "Cancelling"
    CANCELLED = "Cancelled"
    DELETING = "Deleting"
    def __str__(self) -> str:
        return str(self.value)

EngineExecutionOutputV2Type = TypeVar("EngineExecutionOutputV2Type", bound="EngineExecutionOutputV2")

@attr.s(auto_attribs=True)
class EngineExecutionOutputV2(DataContract):
    """Describes the details related to an execution

    """
    executionId: str = None
    outputLocation: str = None
    status: EngineExecutionStatusV2 = None
    message: str = None
    projectId: str = None
    customerId: str = None
    createdAt: str = None
    startedAt: str = None
    updatedAt: str = None
    finishedAt: str = None
    engines: List[EngineDefinitionV2] = None
    poolType: str = None
    virtualMachineSize: str = None
    nodeCount: int = None
    runningSetupIndex: int = None
    runningSetupProgress: str = None
    totalNumberOfSetups: int = None
    maxRunLimitTimeHours: float = None
    scenarioName: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionOutputV2Type, src_dict: Dict[str, Any]) -> EngineExecutionOutputV2Type:
        obj = EngineExecutionOutputV2()
        obj.load_dict(src_dict)
        return obj

EngineInputItemV2Type = TypeVar("EngineInputItemV2Type", bound="EngineInputItemV2")

@attr.s(auto_attribs=True)
class EngineInputItemV2(DataContract):
    """Describes an execution input from a blob storage

    """
    uri: str = None
    localPath: str = None
    engine: str = None
    version: str = None
    runParameters: List[EngineRunParameterV2] = None
    reportLogUpdatesLines: int = None
    logFiles: List[str] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineInputItemV2Type, src_dict: Dict[str, Any]) -> EngineInputItemV2Type:
        obj = EngineInputItemV2()
        obj.load_dict(src_dict)
        return obj

EngineExecutionParametersV2Type = TypeVar("EngineExecutionParametersV2Type", bound="EngineExecutionParametersV2")

@attr.s(auto_attribs=True)
class EngineExecutionParametersV2(DataContract):
    """Describes the parameters used for running an execution with inputs from blob storage

    """
    inputs: List[EngineInputItemV2] = None
    output: EngineOutputItemV2 = None
    platformOutput: List[EngineOutputPlatformItemV2] = None
    options: OptionsV2 = None
    scenarioName: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionParametersV2Type, src_dict: Dict[str, Any]) -> EngineExecutionParametersV2Type:
        obj = EngineExecutionParametersV2()
        obj.load_dict(src_dict)
        return obj

EngineExecutionInputOutputV2Type = TypeVar("EngineExecutionInputOutputV2Type", bound="EngineExecutionInputOutputV2")

@attr.s(auto_attribs=True)
class EngineExecutionInputOutputV2(DataContract):
    platformInputs: EngineExecutionInputParametersV2 = None
    inputs: EngineExecutionParametersV2 = None
    isPlatformData: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionInputOutputV2Type, src_dict: Dict[str, Any]) -> EngineExecutionInputOutputV2Type:
        obj = EngineExecutionInputOutputV2()
        obj.load_dict(src_dict)
        return obj

EngineExecutionSummaryOutputV2Type = TypeVar("EngineExecutionSummaryOutputV2Type", bound="EngineExecutionSummaryOutputV2")

@attr.s(auto_attribs=True)
class EngineExecutionSummaryOutputV2(DataContract):
    """Describes the details shown for an existing execution

    """
    executionId: str = None
    status: EngineExecutionStatusV2 = None
    projectId: str = None
    customerId: str = None
    createdAt: str = None
    startedAt: str = None
    updatedAt: str = None
    finishedAt: str = None
    engines: List[EngineDefinitionV2] = None
    poolType: str = None
    virtualMachineSize: str = None
    nodeCount: int = None
    runningSetupIndex: int = None
    runningSetupProgress: str = None
    totalNumberOfSetups: int = None
    outputLocation: str = None
    maxRunLimitTimeHours: float = None
    scenarioName: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionSummaryOutputV2Type, src_dict: Dict[str, Any]) -> EngineExecutionSummaryOutputV2Type:
        obj = EngineExecutionSummaryOutputV2()
        obj.load_dict(src_dict)
        return obj

EngineExecutionSummaryOutputPagedResultV2Type = TypeVar("EngineExecutionSummaryOutputPagedResultV2Type", bound="EngineExecutionSummaryOutputPagedResultV2")

@attr.s(auto_attribs=True)
class EngineExecutionSummaryOutputPagedResultV2(DataContract):
    _nextLink: str = None
    cursor: str = None
    data: List[EngineExecutionSummaryOutputV2] = None
    __renamed = { "@nextLink": "_nextLink" }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EngineExecutionSummaryOutputPagedResultV2Type, src_dict: Dict[str, Any]) -> EngineExecutionSummaryOutputPagedResultV2Type:
        obj = EngineExecutionSummaryOutputPagedResultV2()
        obj.load_dict(src_dict)
        return obj

class EngineExecutionFieldV2(str, Enum):
    EXECUTIONID = "ExecutionId"
    STATUS = "Status"
    PROJECTID = "ProjectId"
    CREATEDAT = "CreatedAt"
    STARTEDAT = "StartedAt"
    UPDATEDAT = "UpdatedAt"
    FINISHEDAT = "FinishedAt"
    ENGINENAME = "EngineName"
    POOLTYPE = "PoolType"
    NODECOUNT = "NodeCount"
    TOTALNUMBEROFSETUPS = "TotalNumberOfSetups"
    SCENARIONAME = "ScenarioName"
    def __str__(self) -> str:
        return str(self.value)

class EngineGenClientV2(PlatformClient):
    def __init__(self, inspectFnc=PlatformClient.DefaultInspectFnc, **kwargs):
        super().__init__(inspectFnc, includeheaders=PlatformClient.GetServiceHeaders("engine"), **kwargs)

    def GetAllConfigurations(self) -> Response:
        """Get the list of hardware configurations.

        Configuration
        GET /api/compute/configuration/list
        """
        return self.GetRequest("/api/compute/configuration/list", None, api_version="2")

    def GetAllEngines(self) -> Response:
        """Get the list of registered engines/tools.

        Engine
        GET /api/compute/engine/list
        """
        return self.GetRequest("/api/compute/engine/list", None, api_version="2")

    def GetEngine(self, name) -> Response:
        """Get the engine/tool information.

        Engine
        GET /api/compute/engine/{name}
        """
        return self.GetRequest(f"/api/compute/engine/{name}", None, api_version="2")

    def RunExecution(self, projectid, body) -> Response:
        """Run the engine/tool with input data from any blob storage. Create permission on the project is required.

        Execution
        POST /api/compute/execution
        """
        return self.PostRequest("/api/compute/execution", body, None, api_version="2", projectid=projectid)

    def GetExecutions(self, projectid, starttime=None, endtime=None, cursor=None) -> Response:
        """Get information about executions within a project. Read permission on the project is required.

        Execution
        GET /api/compute/execution/list
        """
        queryparams = self.GetQueryParams(startTime=starttime, endTime=endtime, cursor=cursor)
        return self.GetRequest("/api/compute/execution/list", queryparams, api_version="2", projectid=projectid)

    def GetMyExecutions(self, projectid, sortby=None, sortorder=None, starttime=None, endtime=None, scenarioname=None, status=None, project=None, limit=None, cursor=None) -> Response:
        """Get information about executions run by the user.

        Execution
        GET /api/compute/execution/my-list
        """
        queryparams = self.GetQueryParams(sortBy=sortby, sortOrder=sortorder, startTime=starttime, endTime=endtime, scenarioName=scenarioname, status=status, project=project, limit=limit, cursor=cursor)
        return self.GetRequest("/api/compute/execution/my-list", queryparams, api_version="2", projectid=projectid)

    def RunExecutionWithPlatformData(self, projectid, body) -> Response:
        """Run the engine/tool with input data from MIKE Cloud Platform. Create permission on the project is required.

        Execution
        POST /api/compute/execution/platform
        """
        return self.PostRequest("/api/compute/execution/platform", body, None, api_version="2", projectid=projectid, recursivetoken="true")

    def GetExecution(self, projectid, executionid) -> Response:
        """Get the execution information. Read permission on the project is required.

        Execution
        GET /api/compute/execution/{executionId}
        """
        return self.GetRequest(f"/api/compute/execution/{executionid}", None, api_version="2", projectid=projectid)

    def DeleteExecution(self, projectid, executionid) -> Response:
        """Delete the engine/tool execution. It can delete finished (also failed) and running executions. Delete permission on the project is required.

        Execution
        DELETE /api/compute/execution/{executionId}
        """
        return self.DeleteRequest(f"/api/compute/execution/{executionid}", None, api_version="2", projectid=projectid)

    def CancelExecution(self, projectid, executionid) -> Response:
        """Cancel the engine/tool execution. Create permission on the project is required.

        Execution
        PUT /api/compute/execution/{executionId}/cancel
        """
        return self.PutRequest(f"/api/compute/execution/{executionid}/cancel", None, None, api_version="2", projectid=projectid)

    def GetExecutionDiagnostics(self, projectid, executionid) -> Response:
        """Get the diagnostics collected during the execution. Read permission on the project is required.

        Execution
        GET /api/compute/execution/{executionId}/diagnostics
        """
        return self.GetRequest(f"/api/compute/execution/{executionid}/diagnostics", None, api_version="2", projectid=projectid)

    def GetExecutionInputs(self, projectid, executionid) -> Response:
        """Get the inputs used for starting the execution. Read permission on the project is required.

        Execution
        GET /api/compute/execution/{executionId}/input
        """
        return self.GetRequest(f"/api/compute/execution/{executionid}/input", None, api_version="2", projectid=projectid)

    def PutPeekExecutionRequest(self, projectid, body, executionid) -> Response:
        """Peek the engine/tool execution generated files. Read permission on the project is required.

        Execution
        PUT /api/compute/execution/{executionId}/peek
        """
        return self.PutRequest(f"/api/compute/execution/{executionid}/peek", body, None, api_version="2", projectid=projectid)



# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio, json, sys, uuid
from signalr_async.netcore import Hub, Client
from signalr_async.netcore.protocols import JsonProtocol, MessagePackProtocol
from types import TracebackType
from typing import List, Optional, Type
from datetime import datetime

class PubsubSession:
    def __init__(self, clientv2, clientv3, verbose, file=sys.stdout):
        self._clientv2 = clientv2
        self._clientv3 = clientv3
        self._verbose = verbose
        self._file = file
        self._pubsubconnectedevent = asyncio.Event()
        self._client = None
    async def __aenter__(self) -> "PubsubSession":
        await self.start()
        return self
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.stop()
    def print(self, msg):
        print(msg, file=self._file)
    async def wait_pubsubconnected(self):
        await self._pubsubconnectedevent.wait()
    def on_connect(self, sessionid, connectionid):
        if self._verbose > 1:
            self.print(f"CONNECT, {sessionid}, {connectionid}")
        self._pubsubconnectedevent.set()
    def on_reconnect(self, sessionid):
        if self._verbose > 1:
            self.print(f"RECONNECT, {sessionid}")
    def on_disconnect(self, sessionid):
        if self._verbose > 1:
            self.print(f"DISCONNECT, {sessionid}")
    def on_error(self, sessionid):
        self.print(f"ERROR({sessionid})")
        self._finishedevent.set()
    async def on_update(self, event):
        if self._verbose > 1:
            self.print(f"EVENT {json.dumps(event, indent=True)}")
        return True
    async def start(self):
        if self._client:
            await self._client.start()
    async def stop(self):
        if self._client:
            await self._client.stop()
    async def wait(self, timeout: Optional[float] = None):
        if self._client:
            await self._client.wait(timeout)
    def subscribe(self, projectid, sources: List[str], offset: int, offsetpoint: str):
        dhisastoken = self._clientv2.GetSasTokenV2(projectid).Body.get("data")
        huburl = self._clientv3.GetServiceUrlV3("pubsub").Body.get("data")
        if huburl.endswith("/api/"):
            huburl = huburl[0:-5]
        elif huburl.endswith("/api"):
            huburl = huburl[0:-4]
        subscribeurl = f"{huburl}/v1"

        hub = PubsubSubscribeHub(sources, offset, offsetpoint, self)

        headers = { "dhi-sas-token": dhisastoken }
        self._client = Client(subscribeurl, hub,
            connection_options = { #"protocol": MessagePackProtocol(),
                "http_client_options": { "headers": headers }, "ws_client_options": { "headers": headers, "timeout": 1.0 } })

        return self

class PubsubResourceSession(PubsubSession):
    def __init__(self, clientv2, clientv3, verbose, file=sys.stdout):
        super().__init__(clientv2, clientv3, verbose, file)
        self._resourceid = None
        self._resourcereadyevent = asyncio.Event()
    def set_resourceid(self, value):
        self._resourceid = value
        self._resourcereadyevent.set()
    async def on_update(self, event):
        await self._resourcereadyevent.wait()
        if event.get("resourceid") == self._resourceid:
            return await super().on_update(event)
        return False

class ExecutionSession(PubsubResourceSession):
    def __init__(self, showlog, clientv2, clientv3, verbose, file=sys.stdout):
        super().__init__(clientv2, clientv3, verbose, file)
        self._showlog = showlog
        self._resourceid = None
        self._resourcereadyevent = asyncio.Event()
        self._finishedevent = asyncio.Event()
    async def wait_finished(self):
        await self._finishedevent.wait()
    def on_error(self, sessionid):
        super().on_error(sessionid)
        self._finishedevent.set()
    async def on_update(self, event):
        if await super().on_update(event):
            self._showrunprogress(event)
            if self._is_finished_event(event):
                self._finishedevent.set()
    def _is_finished_event(self, event):
        eventtype = event.get("type")
        eventdata = event.get("data")
        if eventtype == "dhi.waterdata.engines.eventing.contracts.messages.v3.engineexecutionstatuschangedmessage" and eventdata:
            eventdatastatus = eventdata.get("Status")
            if eventdatastatus == "Success" and eventdata["RunningSetupIndex"]+1 == eventdata["TotalNumberOfSetups"] \
                or eventdatastatus == "Failure" or eventdatastatus == "Cancelled":
                return True
        return False
    def _showrunprogress(self, event):
        eventtype = event.get("type")
        eventdata = event.get("data")
        msg = None
        if self._showlog:
            if eventtype == "dhi.waterdata.engines.eventing.contracts.messages.v3.engineexecutionlogupdatemessage" and eventdata:
                logfile = eventdata.get("LogFile")
                loglines = eventdata.get("LogLines")
                for line in loglines:
                    self.print(f"{logfile}:{line}")
        if self._verbose > 0:
            if eventtype == "dhi.waterdata.engines.eventing.contracts.messages.v3.engineexecutionstatuschangedmessage" and eventdata:
                status = eventdata.get("Status")
                message = eventdata.get("Message")
                index = eventdata.get("RunningSetupIndex")
                index = index+1 if index != None else ""
                count = eventdata.get("TotalNumberOfSetups")
                msg = f"status={status}, {index}/{count}, {message}" if message else f"status={status}, {index}/{count}"
            elif eventtype == "dhi.waterdata.engines.eventing.contracts.messages.v3.engineexecutionprogressmessage" and eventdata:
                progressinfo = eventdata.get("EngineExecutionProgress")
                if progressinfo:
                    progre = progressinfo.get("PROGRE")
                    index = eventdata.get("RunningSetupIndex")
                    index = index+1 if index != None else ""
                    count = eventdata.get("TotalNumberOfSetups")
                    msg = f"progress={progre}, {index}/{count}"
            else:
                if not eventtype.startswith("dhi.waterdata.engines.eventing.contracts.messages.v2."):
                    msg = f"event={eventtype}"
        if msg:
            self.print(f"{msg}")

class PubsubSubscribeHub(Hub):
    def __init__(self, sources: List[str], offset: int, offsetpoint: str, pubsubsession: PubsubSession):
        super().__init__("subscribe")
        self._subscriptionid = f"dhi-platform-sdk-py-{datetime.now():%Y%m%dT%H%M%S%f}-{uuid.uuid4().hex}"
        self._sources = sources
        self._offset = offset
        self._offsetpoint = offsetpoint
        self._pubsubsession = pubsubsession
    async def on_connect(self, connection_id: str) -> None:
        """Will be awaited after connection established"""
        await self.subscribe()
        if self._pubsubsession:
            self._pubsubsession.on_connect(self._subscriptionid, connection_id)
    async def on_disconnect(self) -> None:
        """Will be awaited after client disconnection"""
        if self._pubsubsession:
            self._pubsubsession.on_disconnect(self._subscriptionid)
    async def on_reconnect(self) -> None:
        if self._pubsubsession:
            self._pubsubsession.on_reconnect(self._subscriptionid)
    async def on_error(self) -> None:
        if self._pubsubsession:
            self._pubsubsession.on_error(self._subscriptionid)
    async def on_Update(self, event) -> None:
        """Invoked by server on (Update)"""
        if self._pubsubsession:
            await self._pubsubsession.on_update(json.loads(event))
    async def subscribe(self) -> bool:
        """Invoke (Subscribe) on server"""
        sub = {
            "SubscriptionId": self._subscriptionid,
            "Filter": { "Sources": self._sources },
            "Condition": { "Offset": self._offset, "OffsetPoint": self._offsetpoint } }
        return await self.invoke("Subscribe", sub)

if __name__ == '__main__':
    print(__file__)
    print(dir())



#!/usr/bin/env python
from dhi.platform.args import ClientArgs
from dhi.platform.config import ClientConfig
from dhi.platform.generated.rawgen import RawGenClientV2
from dhi.platform.rawhelper import RawClientV2Contracts
from dhi.platform.fmt import Format

def __initParser(parser):
    parser.add_argument("-f", "--inputfile", help="Input file path, '-' to read stdin")
    parser.add_argument("--id", action="append", help="Dataset id")
    parser.add_argument("--copyoperationid", action="append", help="Copy operation id")

def main():
    args = ClientArgs.ParseForProject(description="Get copy status", init=__initParser, defaultformat="yaml")
    ClientConfig.UpdateProjectFromConfiguration(args)
    clientv2 = RawGenClientV2(**vars(args))

    input = RawClientV2Contracts.PrepareGetCopyFileStatusInput(
        ClientArgs.LoadJson(args.inputfile),
        id=args.id,
        copyoperationid=args.copyoperationid)
    response = clientv2.GetCopyFilesStatus(args.projectid, input)

    Format.FormatResponse(response, lambda r: r.Body.get("data"), args.format)

if __name__ == '__main__':
    main()



import unittest
import uuid
from dhi.platform.metadata import MetadataClient
from dhi.platform.base.exceptions import MikeCloudRestApiException
from .testcredentials import TEST_IDENTITY

class TestCreateProject(unittest.TestCase):

    _verbosity = 3

    def test_get_nonexisting_project_raises(self):
        metadataclient = MetadataClient(verbose=self._verbosity, identity=TEST_IDENTITY)
        project_id = str(uuid.uuid4())

        try:

            p = metadataclient.get_project(project_id)

        except MikeCloudRestApiException as ex:

            self.assertTrue(ex.message)
            self.assertTrue(ex.status_code)
    
if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python
from dhi.platform.args import ClientArgs
from dhi.platform.config import ClientConfig
from dhi.platform.generated.metadatagen import MetadataGenClientV2
from dhi.platform.fmt import Format

def main():
    args = ClientArgs.ParseForProjectList(description="Get details about deleted project", defaultformat="yaml")
    ClientConfig.UpdatePlatformFromConfiguration(args)
    clientv2 = MetadataGenClientV2(**vars(args))

    responses = (clientv2.GetRecyclableProjectV2(id) for id in args.projectids)

    tablefmt = "{!s:36}\t{!s:16}\t{!s:14}\t{!s:26}\t{!s:26}\t{}"
    tablefields = ["id", "name", "accessLevel", "createdAt", "updatedAt", "deletedAt"]
    Format.FormatResponses(responses, lambda r: r.Body, args.format, tablefmt, tablefields)

if __name__ == '__main__':
    main()
